# Copyright 2026 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Tests for SLR to AST converter."""

import pytest
from pecos.slr import CReg, If, Main, QReg, Repeat
from pecos.slr.ast import (
    AllocatorDecl,
    BinaryOp,
    GateKind,
    GateOp,
    IfStmt,
    MeasureOp,
    PrepareOp,
    Program,
    RegisterDecl,
    RepeatStmt,
    SlrToAst,
    slr_to_ast,
)
from pecos.slr.qalloc import QAlloc
from pecos.slr.qeclib import qubit as qb


class TestSlrToAstBasic:
    """Basic conversion tests."""

    def test_empty_program(self) -> None:
        """Empty program converts to Program with no body."""
        prog = Main()

        ast = slr_to_ast(prog)

        assert isinstance(ast, Program)
        assert ast.name == "Main"
        assert ast.body == ()

    def test_program_with_qreg(self) -> None:
        """QReg converts to AllocatorDecl."""
        prog = Main(
            _q := QReg("q", 2),
        )

        ast = slr_to_ast(prog)

        assert len(ast.declarations) == 1
        decl = ast.declarations[0]
        assert isinstance(decl, AllocatorDecl)
        assert decl.name == "q"
        assert decl.capacity == 2

    def test_program_with_creg(self) -> None:
        """CReg converts to RegisterDecl."""
        prog = Main(
            _c := CReg("c", 3),
        )

        ast = slr_to_ast(prog)

        assert len(ast.declarations) == 1
        decl = ast.declarations[0]
        assert isinstance(decl, RegisterDecl)
        assert decl.name == "c"
        assert decl.size == 3
        assert decl.is_result is True

    def test_program_with_both_regs(self) -> None:
        """Program with QReg and CReg has both declarations."""
        prog = Main(
            _q := QReg("q", 2),
            _c := CReg("c", 2),
        )

        ast = slr_to_ast(prog)

        assert len(ast.declarations) == 2
        names = [d.name for d in ast.declarations]
        assert "q" in names
        assert "c" in names


class TestSlrToAstGates:
    """Gate conversion tests."""

    def test_single_qubit_gate(self) -> None:
        """Single-qubit gate converts to GateOp with correct kind and target."""
        prog = Main(
            q := QReg("q", 1),
            qb.H(q[0]),
        )

        ast = slr_to_ast(prog)

        assert len(ast.body) == 1
        gate = ast.body[0]
        assert isinstance(gate, GateOp)
        assert gate.gate == GateKind.H
        assert len(gate.targets) == 1
        assert gate.targets[0].allocator == "q"
        assert gate.targets[0].index == 0

    def test_two_qubit_gate(self) -> None:
        """Two-qubit gate converts to GateOp with two targets."""
        prog = Main(
            q := QReg("q", 2),
            qb.CX(q[0], q[1]),
        )

        ast = slr_to_ast(prog)

        assert len(ast.body) == 1
        gate = ast.body[0]
        assert isinstance(gate, GateOp)
        assert gate.gate == GateKind.CX
        assert len(gate.targets) == 2
        assert gate.targets[0].index == 0
        assert gate.targets[1].index == 1

    def test_multiple_gates(self) -> None:
        """Multiple gates convert to multiple GateOps in sequence."""
        prog = Main(
            q := QReg("q", 2),
            qb.H(q[0]),
            qb.X(q[1]),
            qb.CZ(q[0], q[1]),
        )

        ast = slr_to_ast(prog)

        assert len(ast.body) == 3
        assert ast.body[0].gate == GateKind.H
        assert ast.body[1].gate == GateKind.X
        assert ast.body[2].gate == GateKind.CZ


class TestSlrToAstPrepMeasure:
    """Prep and Measure conversion tests."""

    def test_prep_operation(self) -> None:
        """Prep converts to PrepareOp with correct allocator and slots."""
        prog = Main(
            q := QReg("q", 2),
            qb.Prep(q[0]),
        )

        ast = slr_to_ast(prog)

        assert len(ast.body) == 1
        prep = ast.body[0]
        assert isinstance(prep, PrepareOp)
        assert prep.allocator == "q"
        assert prep.slots == (0,)

    def test_measure_operation(self) -> None:
        """Measure converts to MeasureOp with targets and results."""
        prog = Main(
            q := QReg("q", 1),
            c := CReg("c", 1),
            qb.Measure(q[0]) > c[0],
        )

        ast = slr_to_ast(prog)

        assert len(ast.body) == 1
        measure = ast.body[0]
        assert isinstance(measure, MeasureOp)
        assert len(measure.targets) == 1
        assert measure.targets[0].allocator == "q"
        assert measure.targets[0].index == 0
        assert len(measure.results) == 1
        assert measure.results[0].register == "c"
        assert measure.results[0].index == 0


class TestSlrToAstControlFlow:
    """Control flow conversion tests."""

    def test_if_statement(self) -> None:
        """If statement converts to IfStmt with then_body."""
        prog = Main(
            q := QReg("q", 1),
            c := CReg("c", 1),
            If(c[0] == 1).Then(
                qb.H(q[0]),
            ),
        )

        ast = slr_to_ast(prog)

        assert len(ast.body) == 1
        if_stmt = ast.body[0]
        assert isinstance(if_stmt, IfStmt)
        assert len(if_stmt.then_body) == 1
        assert isinstance(if_stmt.then_body[0], GateOp)

    def test_if_else_statement(self) -> None:
        """If-else statement converts to IfStmt with both branches."""
        prog = Main(
            q := QReg("q", 1),
            c := CReg("c", 1),
            If(c[0] == 1)
            .Then(
                qb.H(q[0]),
            )
            .Else(
                qb.X(q[0]),
            ),
        )

        ast = slr_to_ast(prog)

        if_stmt = ast.body[0]
        assert isinstance(if_stmt, IfStmt)
        assert len(if_stmt.then_body) == 1
        assert len(if_stmt.else_body) == 1
        assert if_stmt.then_body[0].gate == GateKind.H
        assert if_stmt.else_body[0].gate == GateKind.X

    def test_condition_expression(self) -> None:
        """Condition converts to BinaryExpr with correct operator."""
        prog = Main(
            q := QReg("q", 1),
            c := CReg("c", 1),
            If(c[0] == 1).Then(
                qb.H(q[0]),
            ),
        )

        ast = slr_to_ast(prog)

        if_stmt = ast.body[0]
        # Check the condition is a binary expression
        from pecos.slr.ast import BinaryExpr

        assert isinstance(if_stmt.condition, BinaryExpr)
        assert if_stmt.condition.op == BinaryOp.EQ

    def test_repeat_statement(self) -> None:
        """Repeat statement converts to RepeatStmt with count and body."""
        prog = Main(
            q := QReg("q", 1),
            Repeat(cond=5).block(
                qb.H(q[0]),
            ),
        )

        ast = slr_to_ast(prog)

        assert len(ast.body) == 1
        repeat = ast.body[0]
        assert isinstance(repeat, RepeatStmt)
        assert repeat.count == 5
        assert len(repeat.body) == 1


class TestSlrToAstConverter:
    """Tests for SlrToAst converter class."""

    def test_converter_reusable(self) -> None:
        """Converter can be reused for multiple programs."""
        converter = SlrToAst()

        prog1 = Main(
            q := QReg("q", 1),
            qb.H(q[0]),
        )

        prog2 = Main(
            r := QReg("r", 2),
            qb.X(r[0]),
            qb.X(r[1]),
        )

        ast1 = converter.convert(prog1)
        ast2 = converter.convert(prog2)

        assert len(ast1.body) == 1
        assert len(ast2.body) == 2
        assert ast1.declarations[0].name == "q"
        assert ast2.declarations[0].name == "r"


class TestSlrToAstQEC:
    """Tests for QEC-related patterns."""

    def test_syndrome_extraction_pattern(self) -> None:
        """Syndrome extraction converts with correct operations."""
        prog = Main(
            data := QReg("data", 2),
            ancilla := QReg("ancilla", 1),
            c := CReg("c", 1),
            # Initialize
            qb.Prep(data[0]),
            qb.Prep(data[1]),
            qb.Prep(ancilla[0]),
            # Syndrome extraction
            qb.CX(data[0], ancilla[0]),
            qb.CX(data[1], ancilla[0]),
            qb.Measure(ancilla[0]) > c[0],
        )

        ast = slr_to_ast(prog)

        # Check declarations
        assert len(ast.declarations) == 3
        alloc_names = [d.name for d in ast.declarations if isinstance(d, AllocatorDecl)]
        assert "data" in alloc_names
        assert "ancilla" in alloc_names

        # Check body operations
        assert len(ast.body) == 6
        preps = [op for op in ast.body if isinstance(op, PrepareOp)]
        gates = [op for op in ast.body if isinstance(op, GateOp)]
        measures = [op for op in ast.body if isinstance(op, MeasureOp)]

        assert len(preps) == 3
        assert len(gates) == 2
        assert len(measures) == 1

    def test_round_trip_preserves_structure(self) -> None:
        """Test that conversion preserves the logical structure."""
        prog = Main(
            q := QReg("q", 3),
            c := CReg("c", 1),
            qb.Prep(q[0]),
            qb.Prep(q[1]),
            qb.Prep(q[2]),
            qb.H(q[0]),
            qb.CX(q[0], q[1]),
            qb.CX(q[1], q[2]),
            If(c[0] == 1).Then(
                qb.X(q[0]),
            ),
        )

        ast = slr_to_ast(prog)

        # Verify structure
        assert ast.name == "Main"
        assert len(ast.declarations) == 2

        # Count operation types
        gate_count = sum(1 for op in ast.body if isinstance(op, GateOp))
        prep_count = sum(1 for op in ast.body if isinstance(op, PrepareOp))
        if_count = sum(1 for op in ast.body if isinstance(op, IfStmt))

        assert gate_count == 3  # H, CX, CX
        assert prep_count == 3  # 3 Preps
        assert if_count == 1  # 1 If


class TestSlrToAstQAlloc:
    """Tests for QAlloc (hierarchical allocator) conversion."""

    def test_single_qalloc(self) -> None:
        """Test conversion of a single QAlloc."""
        prog = Main(
            base := QAlloc(4, name="base"),
            base.prepare_all(),
            qb.H(base[0]),
        )

        ast = slr_to_ast(prog)

        assert len(ast.declarations) == 1
        decl = ast.declarations[0]
        assert isinstance(decl, AllocatorDecl)
        assert decl.name == "base"
        assert decl.capacity == 4
        assert decl.parent is None

    def test_hierarchical_qalloc(self) -> None:
        """Test conversion of hierarchical QAllocs (parent-child)."""
        prog = Main(
            base := QAlloc(10, name="base"),
            data := base.child(4, name="data"),
            ancilla := base.child(4, name="ancilla"),
            data.prepare_all(),
            ancilla.prepare_all(),
            qb.H(data[0]),
            qb.CX(data[0], ancilla[0]),
        )

        ast = slr_to_ast(prog)

        # Should have 3 declarations
        assert len(ast.declarations) == 3
        decl_by_name = {d.name: d for d in ast.declarations}

        # Check base allocator
        assert "base" in decl_by_name
        assert decl_by_name["base"].capacity == 10
        assert decl_by_name["base"].parent is None

        # Check child allocators
        assert "data" in decl_by_name
        assert decl_by_name["data"].capacity == 4
        assert decl_by_name["data"].parent == "base"

        assert "ancilla" in decl_by_name
        assert decl_by_name["ancilla"].capacity == 4
        assert decl_by_name["ancilla"].parent == "base"

    def test_qalloc_gates_use_correct_allocator_names(self) -> None:
        """Test that gates on QAlloc qubits reference correct allocator names."""
        prog = Main(
            base := QAlloc(10, name="base"),
            data := base.child(4, name="data"),
            data.prepare_all(),
            qb.H(data[0]),
            qb.CX(data[0], data[1]),
        )

        ast = slr_to_ast(prog)

        # Check that gate targets reference "data" allocator
        assert len(ast.body) == 2
        h_gate = ast.body[0]
        cx_gate = ast.body[1]

        assert h_gate.gate == GateKind.H
        assert h_gate.targets[0].allocator == "data"
        assert h_gate.targets[0].index == 0

        assert cx_gate.gate == GateKind.CX
        assert cx_gate.targets[0].allocator == "data"
        assert cx_gate.targets[1].allocator == "data"

    def test_qalloc_no_duplicates(self) -> None:
        """Test that QAllocs appearing in both vars and ops aren't duplicated."""
        # When using walrus operator, QAllocs end up in ops
        prog = Main(
            base := QAlloc(4, name="base"),
            base.prepare_all(),
            qb.H(base[0]),
        )

        ast = slr_to_ast(prog)

        # Should have exactly 1 declaration, not duplicated
        assert len(ast.declarations) == 1
        names = [d.name for d in ast.declarations]
        assert names.count("base") == 1



"""Test Y and Z gates specifically."""

from guppylang import guppy
from guppylang.std.quantum import measure, qubit, x, y, z
from pecos import Guppy, sim
from pecos_rslib import state_vector


def test_y_gate_only() -> None:
    """Test Y gate by itself."""

    @guppy
    def y_only() -> bool:
        q = qubit()
        y(q)
        return measure(q)

    results = sim(Guppy(y_only)).qubits(1).quantum(state_vector()).run(5).to_dict()
    # measurements is list of lists like [[1], [1], ...], extract last value from each shot
    raw_measurements = results.get("measurements", [])
    measurements = [m[-1] if isinstance(m, list) else m for m in raw_measurements]
    assert all(val == 1 for val in measurements)  # Y|0⟩ should give |1⟩


def test_z_gate_only() -> None:
    """Test Z gate by itself."""

    @guppy
    def z_only() -> bool:
        q = qubit()
        z(q)
        return measure(q)

    results = sim(Guppy(z_only)).qubits(1).quantum(state_vector()).run(5).to_dict()
    # measurements is list of lists like [[0], [0], ...], extract last value from each shot
    raw_measurements = results.get("measurements", [])
    measurements = [m[-1] if isinstance(m, list) else m for m in raw_measurements]
    assert all(val == 0 for val in measurements)  # Z|0⟩ should give |0⟩


def test_y_and_z_tuple() -> None:
    """Test Y and Z gates returning a tuple."""

    @guppy
    def yz_tuple() -> tuple[bool, bool]:
        q1 = qubit()
        y(q1)  # Y|0⟩ = i|1⟩
        r1 = measure(q1)

        q2 = qubit()
        z(q2)  # Z|0⟩ = |0⟩
        r2 = measure(q2)

        return r1, r2

    results = sim(Guppy(yz_tuple)).qubits(2).quantum(state_vector()).run(5).to_dict()
    # measurements is list of lists like [[1, 0], [1, 0], ...] for tuple returns
    raw_measurements = results.get("measurements", [])

    for i in range(5):
        assert raw_measurements[i][0] == 1  # Y|0⟩ should give |1⟩
        assert raw_measurements[i][1] == 0  # Z|0⟩ should give |0⟩


def test_xyz_tuple() -> None:
    """Test X, Y, Z gates returning a tuple."""

    @guppy
    def xyz_tuple() -> tuple[bool, bool, bool]:
        q1 = qubit()
        x(q1)  # X|0⟩ = |1⟩
        r1 = measure(q1)

        q2 = qubit()
        y(q2)  # Y|0⟩ = i|1⟩
        r2 = measure(q2)

        q3 = qubit()
        z(q3)  # Z|0⟩ = |0⟩
        r3 = measure(q3)

        return r1, r2, r3

    results = sim(Guppy(xyz_tuple)).qubits(3).quantum(state_vector()).run(5).to_dict()
    # measurements is list of lists like [[1, 1, 0], [1, 1, 0], ...] for tuple returns
    raw_measurements = results.get("measurements", [])

    for i in range(5):
        assert raw_measurements[i][0] == 1  # X|0⟩ should give |1⟩
        assert raw_measurements[i][1] == 1  # Y|0⟩ should give |1⟩
        assert raw_measurements[i][2] == 0  # Z|0⟩ should give |0⟩



"""Depolarizing error model implementation.

This module provides depolarizing error models for quantum error correction
simulations, including single and two-qubit depolarizing channels with
configurable error rates and gate-specific noise characteristics.
"""

# Copyright 2023 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

from typing import TYPE_CHECKING

import pecos as pc
from pecos.error_models.noise_impl.gate_groups import one_qubits, two_qubits
from pecos.error_models.noise_impl.noise_initz_bitflip import noise_initz_bitflip
from pecos.error_models.noise_impl.noise_meas_bitflip import noise_meas_bitflip
from pecos.error_models.noise_impl.noise_sq_depolarizing import noise_sq_depolarizing
from pecos.error_models.noise_impl.noise_tq_depolarizing import noise_tq_depolarizing

if TYPE_CHECKING:
    from collections.abc import Callable

    from pecos.protocols import MachineProtocol
    from pecos.reps.pyphir.block_types import SeqBlock
    from pecos.reps.pyphir.op_types import QOp

two_qubit_paulis = {
    "IX",
    "IY",
    "IZ",
    "XI",
    "XX",
    "XY",
    "XZ",
    "YI",
    "YX",
    "YY",
    "YZ",
    "ZI",
    "ZX",
    "ZY",
    "ZZ",
}
SYMMETRIC_P2_PAULI_MODEL = dict.fromkeys(two_qubit_paulis, 1 / 15)

one_qubit_paulis = {
    "X",
    "Y",
    "Z",
}
SYMMETRIC_P1_PAULI_MODEL = dict.fromkeys(one_qubit_paulis, 1 / 3)


class DepolarizingErrorModel:
    """Parameterized error mode."""

    def __init__(self, error_params: dict) -> None:
        """Initialize a depolarizing error model.

        Args:
            error_params: Dictionary containing error parameters including:
                - p1: Single-qubit gate error probability
                - p2: Two-qubit gate error probability
                - p_meas: Measurement error probability
                - p_init: Initialization error probability
                - scale: Optional scaling factor for all error rates
                - p1_error_model: Optional custom single-qubit Pauli error distribution
                - p2_error_model: Optional custom two-qubit Pauli error distribution
        """
        self.error_params = dict(error_params)
        self.machine = None
        self.num_qubits = None
        self._eparams: dict | None = None

    def reset(self) -> DepolarizingErrorModel:
        """Reset error generator for another round of syndrome extraction."""
        return DepolarizingErrorModel(error_params=self.error_params)

    def init(self, num_qubits: int, machine: MachineProtocol | None = None) -> None:
        """Initialize the depolarizing error model.

        Args:
            num_qubits: Number of qubits in the system.
            machine: Optional machine protocol for hardware-specific behavior.
        """
        self.machine = machine
        self.num_qubits = num_qubits

        if not self.error_params:
            msg = "Error params not set!"
            raise Exception(msg)

        self._eparams: dict = dict(self.error_params)
        self._scale()

        if "p1_error_model" not in self._eparams:
            self._eparams["p1_error_model"] = SYMMETRIC_P1_PAULI_MODEL

        if "p2_error_model" not in self._eparams:
            self._eparams["p2_error_model"] = SYMMETRIC_P2_PAULI_MODEL

        if "p2_mem" in self._eparams and "p2_mem_error_model" not in self._eparams:
            self._eparams["p2_mem_error_model"] = SYMMETRIC_P2_PAULI_MODEL

    def _scale(self) -> None:
        # conversion from average error to total error
        self._eparams["p1"] *= 3 / 2
        self._eparams["p2"] *= 5 / 4

        scale = self._eparams.get("scale", 1.0)
        self._eparams["p1"] *= scale
        self._eparams["p2"] *= scale

        if isinstance(self._eparams["p_meas"], tuple):
            self._eparams["p_meas"] = pc.mean(self._eparams["p_meas"])

        self._eparams["p_meas"] *= scale
        self._eparams["p_init"] *= scale

    def shot_reinit(self) -> None:
        """Run all code needed at the beginning of each shot, e.g., resetting state."""

    def process(
        self,
        qops: list[QOp],
        _call_back: Callable | None = None,
    ) -> list[QOp | SeqBlock]:
        """Process quantum operations and apply depolarizing errors.

        Args:
            qops: List of quantum operations to process.
            call_back: Optional callback function for additional processing.

        Returns:
            List of quantum operations with applied errors.
        """
        noisy_ops = []

        for op in qops:
            qops_after = None
            qops_before = None
            erroneous_ops = None

            # ########################################
            # INITS WITH X NOISE
            if op.metadata.get("noiseless"):
                pass

            elif op.name in {"init |0>", "Init", "Init +Z"}:
                qops_after = noise_initz_bitflip(
                    op,
                    p=self._eparams["p_init"],
                )

            # ########################################
            # ONE QUBIT GATES
            elif op.name in one_qubits:
                erroneous_ops = noise_sq_depolarizing(
                    op,
                    p=self._eparams["p1"],
                    noise_dict=self._eparams["p1_error_model"],
                )

            # ########################################
            # TWO QUBIT GATES
            elif op.name in two_qubits:
                qops_after = noise_tq_depolarizing(
                    op,
                    p=self._eparams["p2"],
                    noise_dict=self._eparams["p2_error_model"],
                )

                if self._eparams.get("p2_mem"):
                    qops_mem = noise_tq_depolarizing(
                        op,
                        p=self._eparams["p2_mem"],
                        noise_dict=self._eparams["p2_mem_error_model"],
                    )

                    if qops_after:
                        qops_after = qops_after.extend(qops_mem)

            # ########################################
            # MEASURE X NOISE
            elif op.name in {"measure Z", "Measure", "Measure +Z"}:
                erroneous_ops = noise_meas_bitflip(
                    op,
                    p=self._eparams["p_meas"],
                )

            elif op.name in {"Transport", "Idle"}:
                # TODO: Add optional noise model for transport and idle
                erroneous_ops = []

            else:
                msg = f"This error model doesn't handle gate: {op.name}!"
                raise Exception(msg)

            if qops_before:
                noisy_ops.extend(qops_before)

            if erroneous_ops is None:
                noisy_ops.append(op)
            else:
                noisy_ops.extend(erroneous_ops)

            if qops_after:
                noisy_ops.extend(qops_after)

        return noisy_ops



# Copyright 2018 The PECOS Developers
# Copyright 2018 National Technology & Engineering Solutions of Sandia, LLC (NTESS). Under the terms of Contract
# DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Single-qubit commands for sparse stabilizer simulator.

This module provides single-qubit Clifford gate operations for the sparse stabilizer simulator, including Pauli
gates, Hadamard gates, and other Clifford operations optimized for sparse stabilizer tableau representation.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pecos.simulators.sparsesim.state import SparseSim
    from pecos.typing import SimulatorGateParams


def Identity(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    """Identity, which does nothing.

    state (SparseSim): Instance representing the stabilizer state.
    qubit (int): Integer that indexes the qubit being acted on.

    Returns: None

    """


def X(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    """Apply Pauli X gate.

    X -> X
    Z -> -Z
    W -> -W
    Y -> -Y
    => If you have a Z component, add a -1.

    Args:
        state (SparseSim): Instance representing the stabilizer state.
        qubit (int): Integer that indexes the qubit being acted on.
        **_params: Unused additional parameters (kept for interface compatibility).

    Returns: None

    """
    stabs = state.stabs

    # Change the sign appropriately

    # Z -> -1
    # ----------
    stabs.signs_minus ^= stabs.col_z[qubit]


def Y(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    """Pauli Y.

    X -> -X
    Z -> -Z
    W -> W
    Y -> Y
    => If you have an X or Z component but not both, add a -1.

    state (SparseSim): Instance representing the stabilizer state.
    qubit (int): Integer that indexes the qubit being acted on.

    Returns: None

    """
    stabs = state.stabs

    # Change the sign appropriately

    # X or Z (exclusive) -> -1
    # ----------
    stabs.signs_minus ^= stabs.col_x[qubit] ^ stabs.col_z[qubit]


def Z(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    """Apply Pauli Z gate.

    X -> -X
    Z -> Z
    W -> -W
    Y -> -Y
    => If you have a X component, add a -1.

    Args:
        state (SparseSim): Instance representing the stabilizer state.
        qubit (int): Integer that indexes the qubit being acted on.
        **_params: Unused additional parameters (kept for interface compatibility).

    Returns: None

    """
    stabs = state.stabs

    # Change the sign appropriately

    # X -> -1
    # ----------
    stabs.signs_minus ^= stabs.col_x[qubit]


def SX(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    r"""Applies a Q rotation to stabilizers and destabilizers.

    Q = \sqrt{X} = HSH

    Q = Q
    X = Q^2
    Q^{\dagger} = Q^3
    I = Q^4

    X -> X
    Z -> -iW = -Y
    W -> -iZ
    Y -> Z

    state (SparseSim): Instance representing the stabilizer state.
    qubit (int): Integer that indexes the qubit being acted on.

    Returns: None

    """
    stabs = state.stabs

    # Change the sign appropriately

    # Z -> -1
    # ---------------------
    stabs.signs_minus ^= stabs.col_z[qubit]

    # Z -> i
    # ---------------------

    # Now we need to deal with the i's ...

    # Generators common to both
    gens_common = stabs.signs_i & stabs.col_z[qubit]

    # Generators only in Z column
    gens_only_z = stabs.col_z[qubit] - stabs.signs_i

    # Generators that are common => i*i = -1
    # => Update the minus signs
    stabs.signs_minus ^= gens_common

    # Remove them from i's
    stabs.signs_i -= gens_common

    # Generators that are only in Z can just be added => i*1 = i
    stabs.signs_i |= gens_only_z

    # Update Paulis
    # -------------------------------------------------------------------
    for g in state.gens:
        # Update column
        # X += Z
        g.col_x[qubit] ^= g.col_z[qubit]

        for i in g.col_z[qubit]:
            g.row_x[i] ^= {qubit}


def SXdg(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    r"""Applies a Q^{\dagger} rotation to stabilizers and destabilizers.

    Qd = \sqrt{X}^{\dagger} = H S^{\dagger}H

    Q = Q
    X = Q^2
    Q^{\dagger} = Q^3
    I = Q^4

    X -> X
    Z -> iW = Y
    W -> iZ
    Y -> -Z

    state (SparseSim): Instance representing the stabilizer state.
    qubit (int): Integer that indexes the qubit being acted on.

    Returns: None

    """
    stabs = state.stabs

    # Change the sign appropriately

    # Z -> i
    # ---------------------

    # For Zs in the qubit column we want to add i to the signs...

    # Deal with the i's ...

    # Generators common to both
    gens_common = stabs.signs_i & stabs.col_z[qubit]

    # Generators only in Z column
    gens_only_z = stabs.col_z[qubit] - stabs.signs_i

    # Generators that are common => i*i = -1
    # => Update the minus signs
    stabs.signs_minus ^= gens_common

    # Remove them from i's
    stabs.signs_i -= gens_common

    # Generators that are only in Z can just be added => i*1 = i
    stabs.signs_i |= gens_only_z

    # Update Paulis
    # -------------------------------------------------------------------
    for g in state.gens:
        # Update column
        # X += Z
        g.col_x[qubit] ^= g.col_z[qubit]

        for i in g.col_z[qubit]:
            g.row_x[i] ^= {qubit}


def SY(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    r"""Applies a R rotation to stabilizers and destabilizers.

    R = \sqrt{XZ} = SQS^{\dagger}

    R = R
    XZ = R^2
    R^{\dagger} = R^3
    I = R^4

    X -> -Z
    Z -> X
    W -> W
    Y -> Y

    state (SparseSim): Instance representing the stabilizer state.
    qubit (int): Integer that indexes the qubit being acted on.

    Returns: None

    """
    stabs = state.stabs

    # Change the sign appropriately

    # X not Z -> -1
    # ---------------------
    stabs.signs_minus ^= stabs.col_x[qubit] - stabs.col_z[qubit]

    for g in state.gens:
        # Swap X and Z for rows
        xonly = g.col_x[qubit] - g.col_z[qubit]

        zonly = g.col_z[qubit] - g.col_x[qubit]

        for i in xonly:
            g.row_x[i].discard(qubit)
            g.row_z[i].add(qubit)

        for i in zonly:
            g.row_z[i].discard(qubit)
            g.row_x[i].add(qubit)

        # Swap X and Z for cols
        g.col_x[qubit], g.col_z[qubit] = g.col_z[qubit], g.col_x[qubit]


def SYdg(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    r"""Applies a R rotation to stabilizers and destabilizers.

    R^{\dagger} = \sqrt{XZ} = SQ^{\dagger}S^{\dagger}

    R = R
    XZ = R^2
    R^{\dagger} = R^3
    I = R^4

    X -> Z
    Z -> -X
    W -> W
    Y -> Y

    state (SparseSim): Instance representing the stabilizer state.
    qubit (int): Integer that indexes the qubit being acted on.

    Returns: None

    """
    stabs = state.stabs

    # Change the sign appropriately

    # Z not X -> -1
    # ---------------------
    stabs.signs_minus ^= stabs.col_z[qubit] - stabs.col_x[qubit]

    # Update Paulis
    # -------------------------------------------------------------------
    for g in state.gens:
        # Swap X and Z for rows
        xonly = g.col_x[qubit] - g.col_z[qubit]

        zonly = g.col_z[qubit] - g.col_x[qubit]

        for i in xonly:
            g.row_x[i].discard(qubit)
            g.row_z[i].add(qubit)

        for i in zonly:
            g.row_z[i].discard(qubit)
            g.row_x[i].add(qubit)

        # Swap X and Z for cols
        g.col_x[qubit], g.col_z[qubit] = g.col_z[qubit], g.col_x[qubit]


def SZ(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    r"""Applies a phase gate (S) rotation to stabilizers and destabilizers.

    S = \sqrt{Z}

    This is P in CHP.

    S = S
    Z = S^2
    S^{\dagger} = S^3
    I = S^4

    X -> iW = Y
    Z -> Z
    W -> iX
    Y -> -X

    state (SparseSim): Instance representing the stabilizer state.
    qubit (int): Integer that indexes the qubit being acted on.

    Returns: None

    """
    stabs = state.stabs

    # Change the sign appropriately

    # X -> i
    # ---------------------
    # i * i = -1
    stabs.signs_minus ^= stabs.signs_i & stabs.col_x[qubit]
    # For each X add an i unless there is already an i there then delete it.
    stabs.signs_i ^= stabs.col_x[qubit]

    # Update Paulis
    # -------------------------------------------------------------------
    for g in state.gens:
        # Update column
        # Z += X
        g.col_z[qubit] ^= g.col_x[qubit]

        # Update row
        for i in g.col_x[qubit]:
            g.row_z[i] ^= {qubit}


def SZdg(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    r"""Applies a Hermitian adjoint phase gate (S^{\dagger}) rotation to stabilizers and destabilizers.

    S = S
    Z = S^2
    S^{\dagger} = S^3
    I = S^4

    X -> -iW = -Y
    Z -> Z
    W -> -iX
    Y -> X

    state (SparseSim): Instance representing the stabilizer state.
    qubit (int): Integer that indexes the qubit being acted on.

    Returns: None

    """
    stabs = state.stabs

    # Change the sign appropriately

    # X -> -1
    # ---------------------
    stabs.signs_minus ^= stabs.col_x[qubit]

    # X -> i
    # ---------------------
    # Now we need to deal with the i's ...

    # Generators common to both
    gens_common = stabs.signs_i & stabs.col_x[qubit]

    # Generators only in Z column
    gens_only_x = stabs.col_x[qubit] - stabs.signs_i

    # Generators that are common => i*i = -1
    # => Update the minus signs
    stabs.signs_minus ^= gens_common

    # Remove them from i's
    stabs.signs_i -= gens_common

    # Generators that are only in X can just be added => i*1 = i
    stabs.signs_i |= gens_only_x

    # Update Paulis
    # -------------------------------------------------------------------
    for g in state.gens:
        # Update column
        # Z += X
        g.col_z[qubit] ^= g.col_x[qubit]

        for i in g.col_x[qubit]:
            g.row_z[i] ^= {qubit}


def H(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    r"""Applies a Hadamard gate (H) rotation to stabilizers and destabilizers.

    Same as H1 in some places in PECOS.

    X + Z

    X -> Z
    Z -> X
    W -> -W
    Y -> -Y

    state (SparseSim): Instance representing the stabilizer state.
    qubit (int): Integer that indexes the qubit being acted on.

    Returns: None

    """
    stabs = state.stabs

    # Change the sign appropriately

    # X and Z -> -1
    # ----------
    stabs.signs_minus ^= stabs.col_x[qubit] & stabs.col_z[qubit]

    # Update Paulis
    # -------------------------------------------------------------------
    for g in state.gens:
        # Swap X and Z for rows
        xonly = g.col_x[qubit] - g.col_z[qubit]

        zonly = g.col_z[qubit] - g.col_x[qubit]

        for i in xonly:
            g.row_x[i].discard(qubit)
            g.row_z[i].add(qubit)

        for i in zonly:
            g.row_z[i].discard(qubit)
            g.row_x[i].add(qubit)

        # Swap X and Z for cols
        g.col_x[qubit], g.col_z[qubit] = g.col_z[qubit], g.col_x[qubit]


def H2(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    r"""Applies a Hadamard gate (H4) rotation to stabilizers and destabilizers.

    X - Z

    X -> -Z
    Z -> -X
    W -> -W
    Y -> -Y

    state (SparseSim): Instance representing the stabilizer state.
    qubit (int): Integer that indexes the qubit being acted on.

    Returns: None

    """
    stabs = state.stabs

    # Change the sign appropriately

    # X or Z (inclusive) -> -1
    # ----------
    stabs.signs_minus ^= stabs.col_x[qubit] | stabs.col_z[qubit]

    # Update Paulis
    # -------------------------------------------------------------------
    for g in state.gens:
        # Swap X and Z for rows
        xonly = g.col_x[qubit] - g.col_z[qubit]

        zonly = g.col_z[qubit] - g.col_x[qubit]

        for i in xonly:
            g.row_x[i].discard(qubit)
            g.row_z[i].add(qubit)

        for i in zonly:
            g.row_z[i].discard(qubit)
            g.row_x[i].add(qubit)

        # Swap X and Z for cols
        g.col_x[qubit], g.col_z[qubit] = g.col_z[qubit], g.col_x[qubit]


def H3(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    r"""Applies a Hadamard gate (H3) rotation to stabilizers and destabilizers.

    Y + X

    X -> iW = Y
    Z -> -Z
    W -> -iX
    Y -> X

    state (SparseSim): Instance representing the stabilizer state.
    qubit (int): Integer that indexes the qubit being acted on.

    Returns: None

    """
    stabs = state.stabs

    # Change the sign appropriately

    # Z -> -1
    # ----------
    stabs.signs_minus ^= stabs.col_z[qubit]

    # X -> i
    # ----------
    # For Xs in the qubit column we want to add i to the signs...

    # Deal with the i's ...

    # Generators common to both
    gens_common = stabs.signs_i & stabs.col_x[qubit]

    # Generators only in Z column
    gens_only_x = stabs.col_x[qubit] - stabs.signs_i

    # Generators that are common => i*i = -1
    # => Update the minus signs
    stabs.signs_minus ^= gens_common

    # Remove them from i's
    stabs.signs_i -= gens_common

    # Generators that are only in Z can just be added => i*1 = i
    stabs.signs_i |= gens_only_x

    # Update Paulis
    # -------------------------------------------------------------------
    for g in state.gens:
        # Update column
        # X += Z
        g.col_z[qubit] ^= g.col_x[qubit]

        for i in g.col_x[qubit]:
            g.row_z[i] ^= {qubit}


def H4(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    r"""Applies a Hadamard gate (H6) rotation to stabilizers and destabilizers.

    Y - X

    X -> -iW = -Y
    Z -> -Z
    W -> iX
    Y -> -X

    state (SparseSim): Instance representing the stabilizer state.
    qubit (int): Integer that indexes the qubit being acted on.

    Returns: None

    """
    stabs = state.stabs

    # Change the sign appropriately

    # X or Z (exclusive) -> -1
    # ----------
    stabs.signs_minus ^= stabs.col_x[qubit] ^ stabs.col_z[qubit]

    # X -> i
    # ----------
    # For Xs in the qubit column we want to add i to the signs...

    # Deal with the i's ...

    # Generators common to both
    gens_common = stabs.signs_i & stabs.col_x[qubit]

    # Generators only in Z column
    gens_only_x = stabs.col_x[qubit] - stabs.signs_i

    # Generators that are common => i*i = -1
    # => Update the minus signs
    stabs.signs_minus ^= gens_common

    # Remove them from i's
    stabs.signs_i -= gens_common

    # Generators that are only in Z can just be added => i*1 = i
    stabs.signs_i |= gens_only_x

    # Update Paulis
    # -------------------------------------------------------------------
    for g in state.gens:
        # Update column
        # X += Z
        g.col_z[qubit] ^= g.col_x[qubit]

        for i in g.col_x[qubit]:
            g.row_z[i] ^= {qubit}


def H5(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    r"""Applies a Hadamard gate (H2) rotation to stabilizers and destabilizers.

    Z + Y

    X -> -X
    Z -> iW = Y
    W -> -iZ
    Y -> Z
    """
    stabs = state.stabs

    # Change the sign appropriately
    # If X apply -1
    # If Z apply i

    # X -> -1
    # ----------
    stabs.signs_minus ^= stabs.col_x[qubit]

    # Z -> i
    # ----------
    # For Zs in the qubit column we want to add i to the signs...

    # Deal with the i's ...

    # Generators common to both
    gens_common = stabs.signs_i & stabs.col_z[qubit]

    # Generators only in Z column
    gens_only_z = stabs.col_z[qubit] - stabs.signs_i

    # Generators that are common => i*i = -1
    # => Update the minus signs
    stabs.signs_minus ^= gens_common

    # Remove them from i's
    stabs.signs_i -= gens_common

    # Generators that are only in Z can just be added => i*1 = i
    stabs.signs_i |= gens_only_z

    # Update Paulis
    # -------------------------------------------------------------------
    for g in state.gens:
        # Update column
        # X += Z
        g.col_x[qubit] ^= g.col_z[qubit]

        for i in g.col_z[qubit]:
            g.row_x[i] ^= {qubit}


def H6(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    r"""Applies a Hadamard gate (H5) rotation to stabilizers and destabilizers.

    Z - Y

    X -> -X
    Z -> -iW = -Y
    W -> iZ
    Y -> -Z

    state (SparseSim): Instance representing the stabilizer state.
    qubit (int): Integer that indexes the qubit being acted on.

    Returns: None

    """
    stabs = state.stabs

    # Change the sign appropriately

    # X or Z (exclusive) -> -1
    # ----------
    stabs.signs_minus ^= stabs.col_x[qubit] ^ stabs.col_z[qubit]

    # Z -> i
    # ----------
    # For Zs in the qubit column we want to add i to the signs...

    # Deal with the i's ...

    # Generators common to both
    gens_common = stabs.signs_i & stabs.col_z[qubit]

    # Generators only in Z column
    gens_only_z = stabs.col_z[qubit] - stabs.signs_i

    # Generators that are common => i*i = -1
    # => Update the minus signs
    stabs.signs_minus ^= gens_common

    # Remove them from i's
    stabs.signs_i -= gens_common

    # Generators that are only in Z can just be added => i*1 = i
    stabs.signs_i |= gens_only_z

    # Update Paulis
    # -------------------------------------------------------------------
    for g in state.gens:
        # Update column
        # X += Z
        g.col_x[qubit] ^= g.col_z[qubit]

        for i in g.col_z[qubit]:
            g.row_x[i] ^= {qubit}


def F(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    r"""Applies a rotation (F1) about a stabilizer octahedron face to stabilizers  and destabilizers.

    X -> iW = Y
    Z -> X
    W -> -iZ
    Y -> Z

    state (SparseSim): Instance representing the stabilizer state.
    qubit (int): Integer that indexes the qubit being acted on.

    Returns: None

    """
    stabs = state.stabs

    # Change the sign appropriately

    # both X and Z -> -1
    # ----------
    stabs.signs_minus ^= stabs.col_x[qubit] & stabs.col_z[qubit]

    # X -> i
    # ----------
    # For Xs in the qubit column we want to add i to the signs...

    # Deal with the i's ...

    # Generators common to both
    gens_common = stabs.signs_i & stabs.col_x[qubit]

    # Generators only in Z column
    gens_only_x = stabs.col_x[qubit] - stabs.signs_i

    # Generators that are common => i*i = -1
    # => Update the minus signs
    stabs.signs_minus ^= gens_common

    # Remove them from i's
    stabs.signs_i -= gens_common

    # Generators that are only in Z can just be added => i*1 = i
    stabs.signs_i |= gens_only_x

    # Update Paulis
    # -------------------------------------------------------------------
    for g in state.gens:
        # Swap X and Z for rows
        xonly = g.col_x[qubit] - g.col_z[qubit]

        zonly = g.col_z[qubit] - g.col_x[qubit]

        xzshared = g.col_x[qubit] & g.col_z[qubit]

        for i in xzshared:
            g.row_x[i].discard(qubit)

        for i in xonly:
            g.row_z[i].add(qubit)

        # Remove only Z
        # Z -> X
        for i in zonly:
            g.row_z[i].discard(qubit)
            g.row_x[i].add(qubit)

        # Swap X and Z for cols
        g.col_x[qubit], g.col_z[qubit] = g.col_z[qubit], g.col_x[qubit]

        # X += Z
        g.col_x[qubit] ^= g.col_z[qubit]


def F2(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    r"""Applies a rotation (F2) about a stabilizer octahedron face to stabilizers and destabilizers.

    X -> -Z
    Z -> iW = Y
    W -> iX
    Y -> -X

    state (SparseSim): Instance representing the stabilizer state.
    qubit (int): Integer that indexes the qubit being acted on.

    Returns: None

    """
    stabs = state.stabs

    # Change the sign appropriately

    # X not Z -> -1
    # ----------
    stabs.signs_minus ^= stabs.col_x[qubit] - stabs.col_z[qubit]

    # Z -> i
    # ----------
    # For Zs in the qubit column we want to add i to the signs...

    # Deal with the i's ...

    # Generators common to both
    gens_common = stabs.signs_i & stabs.col_z[qubit]

    # Generators only in Z column
    gens_only_z = stabs.col_z[qubit] - stabs.signs_i

    # Generators that are common => i*i = -1
    # => Update the minus signs
    stabs.signs_minus ^= gens_common

    # Remove them from i's
    stabs.signs_i -= gens_common

    # Generators that are only in Z can just be added => i*1 = i
    stabs.signs_i |= gens_only_z

    # Update Paulis
    # -------------------------------------------------------------------
    for g in state.gens:
        # Swap X and Z for rows
        xonly = g.col_x[qubit] - g.col_z[qubit]

        zonly = g.col_z[qubit] - g.col_x[qubit]

        xzshared = g.col_x[qubit] & g.col_z[qubit]

        for i in xzshared:
            g.row_z[i].discard(qubit)

        for i in zonly:
            g.row_x[i].add(qubit)

        # Remove only Z
        # X -> Z
        for i in xonly:
            g.row_x[i].discard(qubit)
            g.row_z[i].add(qubit)

        # Swap X and Z for cols
        g.col_x[qubit], g.col_z[qubit] = g.col_z[qubit], g.col_x[qubit]

        # Z += X
        g.col_z[qubit] ^= g.col_x[qubit]


def F3(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    r"""Applies a rotation (F3) about a stabilizer octahedron face to stabilizers and destabilizers.

    X -> iW = Y
    Z -> -X
    W -> iZ
    Y -> -Z

    state (SparseSim): Instance representing the stabilizer state.
    qubit (int): Integer that indexes the qubit being acted on.

    Returns: None

    """
    stabs = state.stabs

    # Change the sign appropriately

    # Z not X -> -1
    # ----------
    stabs.signs_minus ^= stabs.col_z[qubit] - stabs.col_x[qubit]

    # X -> i
    # ----------
    # For Xs in the qubit column we want to add i to the signs...

    # Deal with the i's ...

    # Generators common to both
    gens_common = stabs.signs_i & stabs.col_x[qubit]

    # Generators only in Z column
    gens_only_x = stabs.col_x[qubit] - stabs.signs_i

    # Generators that are common => i*i = -1
    # => Update the minus signs
    stabs.signs_minus ^= gens_common

    # Remove them from i's
    stabs.signs_i -= gens_common

    # Generators that are only in Z can just be added => i*1 = i
    stabs.signs_i |= gens_only_x

    # Update Paulis
    # -------------------------------------------------------------------
    for g in state.gens:
        # Swap X and Z for rows
        xonly = g.col_x[qubit] - g.col_z[qubit]

        zonly = g.col_z[qubit] - g.col_x[qubit]

        xzshared = g.col_x[qubit] & g.col_z[qubit]

        for i in xzshared:
            g.row_x[i].discard(qubit)

        for i in xonly:
            g.row_z[i].add(qubit)

        # Remove only Z
        # Z -> X
        for i in zonly:
            g.row_z[i].discard(qubit)
            g.row_x[i].add(qubit)

        # Swap X and Z for cols
        g.col_x[qubit], g.col_z[qubit] = g.col_z[qubit], g.col_x[qubit]

        # X += Z
        g.col_x[qubit] ^= g.col_z[qubit]


def F4(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    r"""Applies a rotation (F4) about a stabilizer octahedron face to stabilizers and destabilizers.

    X -> Z
    Z -> -iW = -Y
    W -> iX
    Y -> -X

    state (SparseSim): Instance representing the stabilizer state.
    qubit (int): Integer that indexes the qubit being acted on.

    Returns: None

    """
    stabs = state.stabs

    # Change the sign appropriately

    # Z not X -> -1
    # ----------
    stabs.signs_minus ^= stabs.col_z[qubit] - stabs.col_x[qubit]

    # Z -> i
    # ----------
    # For Zs in the qubit column we want to add i to the signs...

    # Deal with the i's ...

    # Generators common to both
    gens_common = stabs.signs_i & stabs.col_z[qubit]

    # Generators only in Z column
    gens_only_z = stabs.col_z[qubit] - stabs.signs_i

    # Generators that are common => i*i = -1
    # => Update the minus signs
    stabs.signs_minus ^= gens_common

    # Remove them from i's
    stabs.signs_i -= gens_common

    # Generators that are only in Z can just be added => i*1 = i
    stabs.signs_i |= gens_only_z

    # Update Paulis
    # -------------------------------------------------------------------
    for g in state.gens:
        # Swap X and Z for rows
        xonly = g.col_x[qubit] - g.col_z[qubit]

        zonly = g.col_z[qubit] - g.col_x[qubit]

        xzshared = g.col_x[qubit] & g.col_z[qubit]

        for i in xzshared:
            g.row_z[i].discard(qubit)

        for i in zonly:
            g.row_x[i].add(qubit)

        # Remove only Z
        # X -> Z
        for i in xonly:
            g.row_x[i].discard(qubit)
            g.row_z[i].add(qubit)

        # Swap X and Z for cols
        g.col_x[qubit], g.col_z[qubit] = g.col_z[qubit], g.col_x[qubit]

        # Z += X
        g.col_z[qubit] ^= g.col_x[qubit]


def Fdg(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    r"""Applies a rotation (F1^{\dagger}) about a stabilizer octahedron face to stabilizers and destabilizers.

    X -> Z
    Z -> iW = Y
    W -> -iX
    Y -> X

    state (SparseSim): Instance representing the stabilizer state.
    qubit (int): Integer that indexes the qubit being acted on.

    Returns: None

    """
    stabs = state.stabs

    # Change the sign appropriately

    # X and Z -> -1
    # ----------
    stabs.signs_minus ^= stabs.col_x[qubit] & stabs.col_z[qubit]

    # Z -> i
    # ----------
    # For Zs in the qubit column we want to add i to the signs...

    # Deal with the i's ...

    # Generators common to both
    gens_common = stabs.signs_i & stabs.col_z[qubit]

    # Generators only in Z column
    gens_only_z = stabs.col_z[qubit] - stabs.signs_i

    # Generators that are common => i*i = -1
    # => Update the minus signs
    stabs.signs_minus ^= gens_common

    # Remove them from i's
    stabs.signs_i -= gens_common

    # Generators that are only in Z can just be added => i*1 = i
    stabs.signs_i |= gens_only_z

    # Update Paulis
    # -------------------------------------------------------------------
    for g in state.gens:
        # Swap X and Z for rows
        xonly = g.col_x[qubit] - g.col_z[qubit]

        zonly = g.col_z[qubit] - g.col_x[qubit]

        xzshared = g.col_x[qubit] & g.col_z[qubit]

        for i in xzshared:
            g.row_z[i].discard(qubit)

        for i in zonly:
            g.row_x[i].add(qubit)

        # Remove only Z
        # X -> Z
        for i in xonly:
            g.row_x[i].discard(qubit)
            g.row_z[i].add(qubit)

        # Swap X and Z for cols
        g.col_x[qubit], g.col_z[qubit] = g.col_z[qubit], g.col_x[qubit]

        # Z += X
        g.col_z[qubit] ^= g.col_x[qubit]


def F2dg(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    r"""Applies a rotation (F2^{\dagger}) about a stabilizer octahedron face to stabilizers and destabilizers.

    X -> -iW = -Y
    Z -> -X
    W -> -iZ
    Y -> Z

    state (SparseSim): Instance representing the stabilizer state.
    qubit (int): Integer that indexes the qubit being acted on.

    Returns: None

    """
    stabs = state.stabs

    # Change the sign appropriately

    # X or Z (inclusive) -> -1
    # ----------
    stabs.signs_minus ^= stabs.col_x[qubit] | stabs.col_z[qubit]

    # X -> i
    # ----------
    # For Xs in the qubit column we want to add i to the signs...

    # Deal with the i's ...

    # Generators common to both
    gens_common = stabs.signs_i & stabs.col_x[qubit]

    # Generators only in Z column
    gens_only_x = stabs.col_x[qubit] - stabs.signs_i

    # Generators that are common => i*i = -1
    # => Update the minus signs
    stabs.signs_minus ^= gens_common

    # Remove them from i's
    stabs.signs_i -= gens_common

    # Generators that are only in Z can just be added => i*1 = i
    stabs.signs_i |= gens_only_x

    # Update Paulis
    # -------------------------------------------------------------------
    for g in state.gens:
        # Swap X and Z for rows
        xonly = g.col_x[qubit] - g.col_z[qubit]

        zonly = g.col_z[qubit] - g.col_x[qubit]

        xzshared = g.col_x[qubit] & g.col_z[qubit]

        for i in xzshared:
            g.row_x[i].discard(qubit)

        for i in xonly:
            g.row_z[i].add(qubit)

        # Remove only Z
        # Z -> X
        for i in zonly:
            g.row_z[i].discard(qubit)
            g.row_x[i].add(qubit)

        # Swap X and Z for cols
        g.col_x[qubit], g.col_z[qubit] = g.col_z[qubit], g.col_x[qubit]

        # X += Z
        g.col_x[qubit] ^= g.col_z[qubit]


def F3dg(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    r"""Applies a rotation (F3^{\dagger}) about a stabilizer octahedron face to stabilizers and destabilizers.

    X -> -Z
    Z -> -iW = -Y
    W -> -iX
    Y -> X

    state (SparseSim): Instance representing the stabilizer state.
    qubit (int): Integer that indexes the qubit being acted on.

    Returns: None

    """
    stabs = state.stabs

    # Change the sign appropriately

    # X or Z (inclusive) -> -1
    # ----------
    stabs.signs_minus ^= stabs.col_x[qubit] | stabs.col_z[qubit]

    # Z -> i
    # ----------
    # For Zs in the qubit column we want to add i to the signs...

    # Deal with the i's ...

    # Generators common to both
    gens_common = stabs.signs_i & stabs.col_z[qubit]

    # Generators only in Z column
    gens_only_z = stabs.col_z[qubit] - stabs.signs_i

    # Generators that are common => i*i = -1
    # => Update the minus signs
    stabs.signs_minus ^= gens_common

    # Remove them from i's
    stabs.signs_i -= gens_common

    # Generators that are only in Z can just be added => i*1 = i
    stabs.signs_i |= gens_only_z

    # Update Paulis
    # -------------------------------------------------------------------
    for g in state.gens:
        # Swap X and Z for rows
        xonly = g.col_x[qubit] - g.col_z[qubit]

        zonly = g.col_z[qubit] - g.col_x[qubit]

        xzshared = g.col_x[qubit] & g.col_z[qubit]

        for i in xzshared:
            g.row_z[i].discard(qubit)

        for i in zonly:
            g.row_x[i].add(qubit)

        # Remove only Z
        # X -> Z
        for i in xonly:
            g.row_x[i].discard(qubit)
            g.row_z[i].add(qubit)

        # Swap X and Z for cols
        g.col_x[qubit], g.col_z[qubit] = g.col_z[qubit], g.col_x[qubit]

        # Z += X
        g.col_z[qubit] ^= g.col_x[qubit]


def F4dg(state: SparseSim, qubit: int, **_params: SimulatorGateParams) -> None:
    r"""Applies a rotation (F4^{\dagger}) about a stabilizer octahedron face to stabilizers and destabilizers.

    X -> -iW = -Y
    Z -> X
    W -> iZ
    Y -> -Z

    state (SparseSim): Instance representing the stabilizer state.
    qubit (int): Integer that indexes the qubit being acted on.

    Returns: None

    """
    stabs = state.stabs

    # Change the sign appropriately

    # X not Z -> -1
    # ----------
    stabs.signs_minus ^= stabs.col_x[qubit] - stabs.col_z[qubit]

    # X -> i
    # ----------
    # For Xs in the qubit column we want to add i to the signs...

    # Deal with the i's ...

    # Generators common to both
    gens_common = stabs.signs_i & stabs.col_x[qubit]

    # Generators only in Z column
    gens_only_x = stabs.col_x[qubit] - stabs.signs_i

    # Generators that are common => i*i = -1
    # => Update the minus signs
    stabs.signs_minus ^= gens_common

    # Remove them from i's
    stabs.signs_i -= gens_common

    # Generators that are only in Z can just be added => i*1 = i
    stabs.signs_i |= gens_only_x

    # Update Paulis
    # -------------------------------------------------------------------
    for g in state.gens:
        # Swap X and Z for rows
        xonly = g.col_x[qubit] - g.col_z[qubit]

        zonly = g.col_z[qubit] - g.col_x[qubit]

        xzshared = g.col_x[qubit] & g.col_z[qubit]

        for i in xzshared:
            g.row_x[i].discard(qubit)

        for i in xonly:
            g.row_z[i].add(qubit)

        # Remove only Z
        # Z -> X
        for i in zonly:
            g.row_z[i].discard(qubit)
            g.row_x[i].add(qubit)

        # Swap X and Z for cols
        g.col_x[qubit], g.col_z[qubit] = g.col_z[qubit], g.col_x[qubit]

        # X += Z
        g.col_x[qubit] ^= g.col_z[qubit]



# Copyright 2025 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.


"""Tests for BitInt binary integer operations."""

from typing import Final

import pecos as pc
from hypothesis import assume, given
from hypothesis import strategies as st
from pecos import BitInt

# BitInt uses actual fixed-width arithmetic, unlike the original BitInt which used
# Python's arbitrary-precision int internally with i64 dtype. Use 64 bits to match i64 range.
DEFAULT_SIZE: Final = 64
MIN: Final = -(2 ** (DEFAULT_SIZE - 1))  # -2^63 for signed 64-bit
MAX: Final = 2 ** (DEFAULT_SIZE - 1) - 1  # 2^63 - 1 for signed 64-bit
int_range = st.integers(min_value=MIN, max_value=MAX)


@given(st.text(alphabet=["0", "1"], min_size=1))
def test_init(x: str) -> None:
    """Test BitInt initialization from binary string."""
    ba = BitInt(x)
    assert ba == f"0b{x}"


def test_set_bit() -> None:
    """Test setting individual bits in BitInt."""
    ba = BitInt("0000")
    ba[2] = 1
    assert ba == 0b0100


def test_get_bit() -> None:
    """Test getting individual bits from BitInt."""
    ba = BitInt("1010")
    assert ba[2] == 0
    assert ba[3] == 1


def test_to_int() -> None:
    """Test converting BitInt to integer."""
    ba = BitInt("1010")
    assert int(ba) == 10


@given(int_range, int_range)
def test_addition(x: int, y: int) -> None:
    """Test BitInt addition operation."""
    assume(MIN <= x + y <= MAX)
    ba1 = BitInt(DEFAULT_SIZE, x)
    ba2 = BitInt(DEFAULT_SIZE, y)
    result = ba1 + ba2
    assert int(result) == x + y


def test_subtraction() -> None:
    """Test BitInt subtraction operation."""
    ba1 = BitInt("1101")  # 13
    ba2 = BitInt("1010")  # 10
    result = ba1 - ba2
    assert int(result) == 3


@given(int_range, int_range)
def test_multiplication(x: int, y: int) -> None:
    """Test BitInt multiplication operation."""
    assume(MIN <= x * y <= MAX)
    ba1 = BitInt(DEFAULT_SIZE, x)
    ba2 = BitInt(DEFAULT_SIZE, y)
    result = ba1 * ba2
    assert int(result) == x * y


def test_comparison() -> None:
    """Test BitInt comparison operations."""
    ba1 = BitInt("1010")  # 10
    ba2 = BitInt("1010")  # 10
    ba3 = BitInt("1101")  # 13
    assert ba1 == ba2
    assert ba1 != ba3
    assert ba1 != ba3
    assert ba1 < ba3
    assert ba3 > ba1


def test_bitwise_and() -> None:
    """Test BitInt bitwise AND operation."""
    ba1 = BitInt("1010")  # 10
    ba2 = BitInt("1101")  # 13
    result = ba1 & ba2
    assert result == 0b1000


def test_bitwise_or() -> None:
    """Test BitInt bitwise OR operation."""
    ba1 = BitInt("1010")  # 10
    ba2 = BitInt("1101")  # 13
    result = ba1 | ba2
    assert result == 0b1111


def test_bitwise_xor() -> None:
    """Test BitInt bitwise XOR operation."""
    ba1 = BitInt("1010")  # 10
    ba2 = BitInt("1101")  # 13
    result = ba1 ^ ba2
    assert result == 0b0111


def test_unsigned_bitwise_not() -> None:
    """Test BitInt bitwise NOT operation for unsigned data."""
    ba = BitInt("1010", dtype=pc.u64)  # 10
    result = ~ba
    assert result == 0b0101


@given(int_range)
def test_signed_bitwise_not(x: int) -> None:
    """Test BitInt bitwise NOT operation for signed data."""
    ba = BitInt(DEFAULT_SIZE, x)
    result = ~ba
    assert int(result) == -x - 1  # (two's complement)



"""Dependency analyzer for SLR blocks."""

from __future__ import annotations

import inspect
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from pecos.slr import Block


@dataclass
class BlockDependency:
    """Represents dependencies for a block."""

    block_type: str
    constructor_params: dict[str, Any]  # parameter name -> value/type
    used_variables: set[str]  # Set of variable names used in operations
    nested_blocks: list[BlockDependency]  # Dependencies of nested blocks


class DependencyAnalyzer:
    """Analyzes SLR blocks to determine parameter dependencies."""

    def __init__(self):
        self.analyzed_blocks = {}  # Cache of analyzed block types

    def analyze_block(self, block: Block) -> BlockDependency:
        """Analyze a block to determine its dependencies."""
        block_type = type(block).__name__

        # Get constructor parameters
        constructor_params = self._get_constructor_params(block)

        # Find used variables in operations
        used_variables = set()
        nested_blocks = []

        if hasattr(block, "ops"):
            for op in block.ops:
                # Collect variables from operations
                self._collect_variables_from_op(op, used_variables)

                # If it's a nested block, analyze it too
                if hasattr(op, "ops") and hasattr(op, "vars"):
                    nested_dep = self.analyze_block(op)
                    nested_blocks.append(nested_dep)
                    # Add nested block's used variables
                    used_variables.update(nested_dep.used_variables)

        return BlockDependency(
            block_type=block_type,
            constructor_params=constructor_params,
            used_variables=used_variables,
            nested_blocks=nested_blocks,
        )

    def _get_constructor_params(self, block: Block) -> dict[str, Any]:
        """Extract constructor parameters from a block instance."""
        params = {}

        # Get the constructor signature
        sig = inspect.signature(type(block).__init__)

        # Try to match parameters with instance attributes
        for param_name in sig.parameters:
            if param_name == "self":
                continue

            # Common patterns for how parameters are stored
            if hasattr(block, param_name):
                params[param_name] = getattr(block, param_name)
            elif hasattr(block, f"_{param_name}"):
                params[param_name] = getattr(block, f"_{param_name}")
            # Try to infer from operations
            elif param_name in ["data", "qubits", "q"]:
                # Look for quantum registers
                params[param_name] = self._find_qreg_in_ops(block)
            elif param_name in ["ancilla", "a"]:
                # Look for ancilla qubits
                params[param_name] = self._find_ancilla_in_ops(block)
            elif param_name in ["init_bit", "init", "bit", "c"]:
                # Look for classical bits
                params[param_name] = self._find_bit_in_ops(block)

        return params

    def _collect_variables_from_op(self, op, used_vars: set[str]):
        """Collect variable names used in an operation."""
        # Check quantum arguments
        if hasattr(op, "qargs"):
            for qarg in op.qargs:
                if hasattr(qarg, "reg") and hasattr(qarg.reg, "sym"):
                    used_vars.add(qarg.reg.sym)
                elif hasattr(qarg, "sym"):
                    # Direct QReg object
                    used_vars.add(qarg.sym)
                elif isinstance(qarg, tuple):
                    # Handle tuples of qubits
                    for q in qarg:
                        if hasattr(q, "reg") and hasattr(q.reg, "sym"):
                            used_vars.add(q.reg.sym)

        # Check classical arguments
        if hasattr(op, "cargs"):
            for carg in op.cargs:
                if hasattr(carg, "reg") and hasattr(carg.reg, "sym"):
                    used_vars.add(carg.reg.sym)

        # Check output bits (for measurements)
        if hasattr(op, "cout") and op.cout:
            for cout in op.cout:
                if hasattr(cout, "reg") and hasattr(cout.reg, "sym"):
                    used_vars.add(cout.reg.sym)
                elif hasattr(cout, "sym"):
                    # Direct CReg reference
                    used_vars.add(cout.sym)

        # Check condition (for If blocks)
        if hasattr(op, "cond"):
            self._collect_variables_from_expr(op.cond, used_vars)

        # Check expressions (for classical operations)
        if hasattr(op, "left"):
            self._collect_variables_from_expr(op.left, used_vars)
        if hasattr(op, "right"):
            self._collect_variables_from_expr(op.right, used_vars)

    def _collect_variables_from_expr(self, expr, used_vars: set[str]):
        """Collect variable names from expressions."""
        if hasattr(expr, "reg") and hasattr(expr.reg, "sym"):
            used_vars.add(expr.reg.sym)
        elif hasattr(expr, "left") and hasattr(expr, "right"):
            # Binary operation
            self._collect_variables_from_expr(expr.left, used_vars)
            self._collect_variables_from_expr(expr.right, used_vars)
        elif hasattr(expr, "args"):
            # Function call or similar
            for arg in expr.args:
                self._collect_variables_from_expr(arg, used_vars)

    def _find_qreg_in_ops(self, block):
        """Try to find quantum register used in operations."""
        if hasattr(block, "ops"):
            for op in block.ops:
                if hasattr(op, "qargs") and op.qargs:
                    qarg = op.qargs[0]
                    if hasattr(qarg, "reg"):
                        return qarg.reg
        return None

    def _find_ancilla_in_ops(self, block):
        """Try to find ancilla qubit used in operations."""
        # Look for single qubit operations that might be ancilla
        if hasattr(block, "ops"):
            for op in block.ops:
                if type(op).__name__ == "Prep" and hasattr(op, "qargs"):
                    # Prep operations often reset ancillas
                    for qarg in op.qargs:
                        if hasattr(qarg, "index") and not hasattr(qarg, "size"):
                            # Single qubit
                            return qarg
        return None

    def _find_bit_in_ops(self, block):
        """Try to find classical bit used in operations."""
        if hasattr(block, "ops"):
            for op in block.ops:
                if hasattr(op, "cout") and op.cout:
                    # Measurement output
                    return op.cout[0]
        return None

    def get_required_parameters(
        self,
        block: Block,
        parent_context: dict[str, Any],
    ) -> list[tuple[str, str]]:
        """Get the parameters required for a block function.

        Args:
            block: The block to analyze
            parent_context: Dictionary mapping variable names to their types/values

        Returns:
            List of (param_name, param_type) tuples
        """
        dep = self.analyze_block(block)

        # Collect all used variables
        all_used = dep.used_variables.copy()

        # Map to parameter types
        params = []
        for var_name in sorted(all_used):
            if var_name in parent_context:
                var_info = parent_context[var_name]
                if hasattr(var_info, "__class__"):
                    var_type = var_info.__class__.__name__
                    if var_type == "QReg":
                        size = var_info.size if hasattr(var_info, "size") else 1
                        params.append((var_name, f"array[quantum.qubit, {size}]"))
                    elif var_type == "CReg":
                        size = var_info.size if hasattr(var_info, "size") else 1
                        params.append((var_name, f"array[bool, {size}]"))
                    else:
                        params.append((var_name, var_type))

        return params



"""Test noise model integration with sim.

This test file verifies that noise models are properly integrated
and working with the sim builder pattern.
"""

import pytest
from guppylang import guppy
from guppylang.std.quantum import cx, h, measure, qubit, x
from pecos import Guppy, sim
from pecos_rslib import (
    biased_depolarizing_noise,
    depolarizing_noise,
    general_noise,
    state_vector,
)


class TestNoiseModels:
    """Test noise model integration with sim."""

    def test_no_noise_deterministic(self) -> None:
        """Test that circuits without noise are deterministic."""

        @guppy
        def deterministic_circuit() -> bool:
            q = qubit()
            x(q)
            return measure(q)

        # Run with seed for reproducibility
        results = sim(deterministic_circuit).qubits(10).quantum(state_vector()).seed(42).run(10).to_dict()

        # Should always measure |1⟩
        raw_measurements = results.get("measurements", [])
        measurements = [m[-1] if isinstance(m, list) else m for m in raw_measurements]
        assert all(r == 1 for r in measurements)

    def test_depolarizing_noise_effect(self) -> None:
        """Test that depolarizing noise introduces errors."""

        @guppy
        def simple_circuit() -> bool:
            q = qubit()
            x(q)
            return measure(q)

        # Create depolarizing noise - must chain all probability setters
        noise = (
            depolarizing_noise()
            .with_prep_probability(0.0)  # No prep errors
            .with_p1_probability(0.2)  # 20% chance of error on single-qubit gates
            .with_p2_probability(0.0)  # No two-qubit gate errors
            .with_meas_probability(0.0)
        )  # No measurement errors

        # High depolarizing probability to see effect
        results = sim(simple_circuit).qubits(10).quantum(state_vector()).noise(noise).seed(42).run(100).to_dict()

        raw_measurements = results.get("measurements", [])
        measurements = [m[-1] if isinstance(m, list) else m for m in raw_measurements]

        # With 0.2 depolarizing on X gate, we should see some 0s
        zeros = sum(1 for r in measurements if r == 0)
        assert zeros > 0, f"Depolarizing noise should introduce errors, got {zeros} zeros"
        assert zeros < 100, "Should not flip all bits"

    def test_biased_depolarizing_noise(self) -> None:
        """Test biased depolarizing noise model."""

        @guppy
        def simple_circuit() -> bool:
            q = qubit()
            x(q)
            return measure(q)

        # Use biased depolarizing - must chain all probability setters
        noise = (
            biased_depolarizing_noise()
            .with_prep_probability(0.05)  # State prep errors
            .with_p1_probability(0.1)  # Single-qubit gate errors
            .with_p2_probability(0.0)  # No two-qubit gate errors
            .with_meas_0_probability(0.05)  # Measurement errors for |0⟩
            .with_meas_1_probability(0.05)
        )  # Measurement errors for |1⟩

        results = sim(simple_circuit).qubits(10).quantum(state_vector()).noise(noise).seed(42).run(100).to_dict()

        raw_measurements = results.get("measurements", [])
        measurements = [m[-1] if isinstance(m, list) else m for m in raw_measurements]

        # Should see some errors
        zeros = sum(1 for r in measurements if r == 0)
        assert zeros > 0, "Biased depolarizing should introduce errors"

    def test_general_noise_model(self) -> None:
        """Test general noise model builder."""

        @guppy
        def simple_circuit() -> bool:
            q = qubit()
            x(q)
            return measure(q)

        # Use general noise model with multiple error types
        noise_builder = (
            general_noise().with_p1_probability(0.01).with_prep_probability(0.01)  # Single-qubit gate errors
        )  # Preparation errors

        results = (
            sim(simple_circuit).qubits(10).quantum(state_vector()).noise(noise_builder).seed(42).run(100).to_dict()
        )

        raw_measurements = results.get("measurements", [])
        measurements = [m[-1] if isinstance(m, list) else m for m in raw_measurements]

        # Should see some errors but not too many
        sum(1 for r in measurements if r == 0)
        # With low error rates, might not see errors in 100 shots
        # Just verify it runs without crashing
        assert len(measurements) == 100

    def test_noise_models_comparison(self) -> None:
        """Compare different noise models on same circuit."""

        @guppy
        def bell_circuit() -> tuple[bool, bool]:
            q1 = qubit()
            q2 = qubit()
            h(q1)
            cx(q1, q2)
            return measure(q1), measure(q2)

        # Run without noise
        results_clean = sim(Guppy(bell_circuit)).qubits(10).quantum(state_vector()).seed(42).run(100).to_dict()

        # Run with depolarizing noise - chain all probability setters
        noise = (
            depolarizing_noise()
            .with_prep_probability(0.0)  # No prep errors
            .with_p1_probability(0.05)  # 5% error on single-qubit gates
            .with_p2_probability(0.05)  # 5% error on two-qubit gates
            .with_meas_probability(0.0)
        )  # No measurement errors

        results_noisy = sim(bell_circuit).qubits(10).quantum(state_vector()).noise(noise).seed(42).run(100).to_dict()

        # Extract measurements - format is [[m0, m1], [m0, m1], ...]
        clean_measurements = results_clean.get("measurements", [])
        noisy_measurements = results_noisy.get("measurements", [])

        # Check correlations
        clean_corr = sum(1 for m in clean_measurements if m[0] == m[1])
        noisy_corr = sum(1 for m in noisy_measurements if m[0] == m[1])

        # Clean Bell state should have perfect correlation
        assert clean_corr == 100, "Clean Bell state should be perfectly correlated"

        # Noisy should have less correlation (or might still be perfect with low noise)
        # Just verify it runs
        assert noisy_corr >= 0


def test_noise_model_builder_pattern() -> None:
    """Test the builder pattern for noise models."""

    @guppy
    def simple_x_circuit() -> bool:
        q = qubit()
        x(q)
        return measure(q)

    # Test that builder pattern works - chain all probability setters
    noise1 = (
        depolarizing_noise()
        .with_prep_probability(0.0)
        .with_p1_probability(0.1)
        .with_p2_probability(0.0)
        .with_meas_probability(0.0)
        .with_seed(1)
    )

    results1 = sim(simple_x_circuit).qubits(10).quantum(state_vector()).noise(noise1).seed(42).run(10).to_dict()

    # Different seed should give different results
    noise2 = (
        depolarizing_noise()
        .with_prep_probability(0.0)
        .with_p1_probability(0.1)
        .with_p2_probability(0.0)
        .with_meas_probability(0.0)
        .with_seed(2)
    )

    results2 = sim(simple_x_circuit).qubits(10).quantum(state_vector()).noise(noise2).seed(43).run(10).to_dict()

    raw_measurements1 = results1.get("measurements", [])
    raw_measurements2 = results2.get("measurements", [])
    measurements1 = [m[-1] if isinstance(m, list) else m for m in raw_measurements1]
    measurements2 = [m[-1] if isinstance(m, list) else m for m in raw_measurements2]

    # With different seeds in noise models, we might get different error patterns
    # But with only 10 shots, they might be the same. Just check they both run.
    assert len(measurements1) == 10
    assert len(measurements2) == 10


def test_noise_on_single_qubit_gates() -> None:
    """Test noise specifically on single-qubit gates."""

    @guppy
    def multi_gate_circuit() -> bool:
        q = qubit()
        h(q)  # Should get noise
        x(q)  # Should get noise
        return measure(q)

    # Configure noise only for single-qubit gates
    noise = general_noise().with_p1_probability(0.3)  # High error rate to see effect

    results = sim(multi_gate_circuit).qubits(10).quantum(state_vector()).noise(noise).seed(42).run(100).to_dict()

    raw_measurements = results.get("measurements", [])
    measurements = [m[-1] if isinstance(m, list) else m for m in raw_measurements]

    # H followed by X should give |1⟩ without noise
    # With noise, we should see some 0s
    zeros = sum(1 for r in measurements if r == 0)
    assert zeros > 0, "Noise on gates should cause errors"


def test_measurement_noise() -> None:
    """Test measurement noise specifically."""

    @guppy
    def simple_circuit() -> bool:
        q = qubit()
        x(q)
        return measure(q)

    # Configure noise only for measurements
    noise = general_noise().with_meas_probability(0.2)  # High measurement error

    results = sim(simple_circuit).qubits(10).quantum(state_vector()).noise(noise).seed(42).run(100).to_dict()

    raw_measurements = results.get("measurements", [])
    measurements = [m[-1] if isinstance(m, list) else m for m in raw_measurements]

    # X gate gives |1⟩, but measurement errors should flip some
    zeros = sum(1 for r in measurements if r == 0)
    assert zeros > 0, "Measurement noise should cause readout errors"



# Copyright 2026 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Tests for validation pipeline."""

import math

from pecos.slr import CReg, If, Main, QReg
from pecos.slr.ast import slr_to_ast
from pecos.slr.ast.nodes import (
    AllocatorDecl,
    GateKind,
    GateOp,
    Program,
    SlotRef,
)
from pecos.slr.ast.validation import (
    BoundsChecker,
    TypeChecker,
    ValidationPipeline,
    ValidationResult,
    create_default_pipeline,
    validate,
)
from pecos.slr.qeclib import qubit as qb


class TestValidateFunction:
    """Tests for the validate() convenience function."""

    def test_validate_valid_program(self) -> None:
        """Valid program passes all checks."""
        prog = Main(
            q := QReg("q", 2),
            qb.H(q[0]),
            qb.CX(q[0], q[1]),
        )

        ast = slr_to_ast(prog)
        result = validate(ast)

        assert result.valid is True
        assert len(result.errors) == 0

    def test_validate_with_rotation(self) -> None:
        """Program with rotation gates passes."""
        prog = Main(
            q := QReg("q", 1),
            qb.RZ[0.5](q[0]),
            qb.RX[math.pi](q[0]),
        )

        ast = slr_to_ast(prog)
        result = validate(ast)

        assert result.valid is True

    def test_validate_complex_circuit(self) -> None:
        """Complex circuit with all features."""
        prog = Main(
            q := QReg("q", 3),
            c := CReg("c", 1),
            qb.H(q[0]),
            qb.CX(q[0], q[1]),
            qb.CX(q[1], q[2]),
            qb.RZ[0.5](q[0]),
            If(c[0] == 1).Then(
                qb.X(q[0]),
            ),
        )

        ast = slr_to_ast(prog)
        result = validate(ast)

        assert result.valid is True


class TestValidationPipeline:
    """Tests for ValidationPipeline class."""

    def test_empty_pipeline(self) -> None:
        """Empty pipeline returns valid result."""
        pipeline = ValidationPipeline([])

        prog = Main(q := QReg("q", 1), qb.H(q[0]))
        ast = slr_to_ast(prog)

        result = pipeline.validate(ast)

        assert result.valid is True
        assert len(result.passes_applied) == 0

    def test_custom_pipeline(self) -> None:
        """Custom pipeline with specific passes."""
        pipeline = ValidationPipeline([BoundsChecker()])

        prog = Main(q := QReg("q", 1), qb.H(q[0]))
        ast = slr_to_ast(prog)

        result = pipeline.validate(ast)

        assert result.valid is True
        assert "bounds_checker" in result.passes_applied

    def test_add_pass(self) -> None:
        """Adding passes to pipeline."""
        pipeline = ValidationPipeline()
        pipeline.add_pass(BoundsChecker())
        pipeline.add_pass(TypeChecker())

        prog = Main(q := QReg("q", 1), qb.H(q[0]))
        ast = slr_to_ast(prog)

        result = pipeline.validate(ast)

        assert result.valid is True
        assert len(result.passes_applied) == 2

    def test_pipeline_accumulates_errors(self) -> None:
        """Pipeline accumulates errors from all passes."""
        pipeline = ValidationPipeline([BoundsChecker(), TypeChecker()])

        # Create program with multiple types of errors
        prog = Program(
            name="test",
            allocator=AllocatorDecl(name="q", capacity=2),
            body=(
                GateOp(
                    gate=GateKind.H,
                    targets=(SlotRef(allocator="q", index=5),),  # Bounds error
                ),
                GateOp(
                    gate=GateKind.H,
                    targets=(
                        SlotRef(allocator="q", index=0),
                        SlotRef(allocator="q", index=1),  # Arity error
                    ),
                ),
            ),
        )

        result = pipeline.validate(prog)

        assert result.valid is False
        assert len(result.errors) >= 2  # At least one from each pass


class TestDefaultPipeline:
    """Tests for create_default_pipeline()."""

    def test_default_pipeline_runs_all(self) -> None:
        """Default pipeline runs all standard passes."""
        pipeline = create_default_pipeline()

        prog = Main(q := QReg("q", 1), qb.H(q[0]))
        ast = slr_to_ast(prog)

        result = pipeline.validate(ast)

        assert result.valid is True
        assert "bounds_checker" in result.passes_applied
        assert "type_checker" in result.passes_applied
        assert "allocation_validator" in result.passes_applied


class TestValidationResult:
    """Tests for ValidationResult class."""

    def test_result_merge(self) -> None:
        """Merging validation results."""
        result1 = ValidationResult(valid=True, passes_applied=["pass1"])
        result2 = ValidationResult(valid=False, errors=[], passes_applied=["pass2"])

        merged = result1.merge(result2)

        assert merged.valid is False
        assert "pass1" in merged.passes_applied
        assert "pass2" in merged.passes_applied

    def test_result_string_valid(self) -> None:
        """String representation for valid result."""
        result = ValidationResult(valid=True)
        assert "Valid" in str(result)

    def test_result_string_invalid(self) -> None:
        """String representation for invalid result."""
        result = ValidationResult(valid=False)
        assert "Invalid" in str(result)

    def test_error_count(self) -> None:
        """Error count property."""
        from pecos.slr.ast.validation import ValidationError

        result = ValidationResult(
            valid=False,
            errors=[
                ValidationError(message="error1"),
                ValidationError(message="error2"),
            ],
        )

        assert result.error_count == 2

    def test_warning_count(self) -> None:
        """Warning count property."""
        from pecos.slr.ast.validation import Severity, ValidationError

        result = ValidationResult(
            valid=True,
            warnings=[
                ValidationError(message="warn1", severity=Severity.WARNING),
            ],
        )

        assert result.warning_count == 1



"""Test the IR-based Guppy generator."""

from pecos.slr import CReg, If, Main, QReg
from pecos.slr.gen_codes.guppy.ir_generator import IRGuppyGenerator
from pecos.slr.qeclib import qubit
from pecos.slr.qeclib.qubit.measures import Measure


def test_ir_simple_measurement() -> None:
    """Test simple measurement with IR generator."""
    prog = Main(
        q := QReg("q", 2),
        c := CReg("c", 2),
        Measure(q[0]) > c[0],
        Measure(q[1]) > c[1],
    )
    gen = IRGuppyGenerator(_internal=True)
    gen.generate_block(prog)
    code = gen.get_output()

    # Check basic structure
    assert "@guppy" in code
    assert "def main() -> None:" in code
    # With optimization, q might be dynamically allocated instead of pre-allocated
    assert "q = array(quantum.qubit() for _ in range(2))" in code or "q_0 = quantum.qubit()" in code
    assert "c = array(False for _ in range(2))" in code

    # Should have measurements - format depends on allocation strategy
    # Array indexing for pre-allocated, unpacked variables for dynamic
    if "q_0 = quantum.qubit()" in code:
        # Dynamic allocation uses unpacked variables
        assert "c_0 = quantum.measure(q_0)" in code
        assert "c_1 = quantum.measure(q_1)" in code
    else:
        # Pre-allocation uses array indexing
        assert "c[0] = quantum.measure(q[0])" in code
        assert "c[1] = quantum.measure(q[1])" in code
    assert 'result("c", c)' in code


def test_ir_full_array_measurement() -> None:
    """Test full array measurement with IR generator."""
    prog = Main(
        q := QReg("q", 3),
        c := CReg("c", 3),
        Measure(q) > c,
    )
    gen = IRGuppyGenerator(_internal=True)
    gen.generate_block(prog)
    code = gen.get_output()

    # Should use measure_array for full measurement
    assert "c = quantum.measure_array(q)" in code
    # Should NOT unpack
    assert "=" not in code or "= q" not in code or ", q_" not in code


def test_ir_quantum_gates() -> None:
    """Test quantum gates with IR generator."""
    prog = Main(
        q := QReg("q", 2),
        c := CReg("c", 2),
        qubit.H(q[0]),
        qubit.CX(q[0], q[1]),
        Measure(q) > c,
    )
    gen = IRGuppyGenerator(_internal=True)
    gen.generate_block(prog)
    code = gen.get_output()

    # Should use array indexing
    assert "quantum.h(q[0])" in code
    assert "quantum.cx(q[0], q[1])" in code


def test_ir_conditional_resources() -> None:
    """Test conditional resource consumption with IR generator."""
    prog = Main(
        q := QReg("q", 2),
        c := CReg("c", 2),
        flag := CReg("flag", 1),
        Measure(q[0]) > flag[0],
        If(flag[0]).Then(
            Measure(q[1]) > c[1],
        ),
    )
    gen = IRGuppyGenerator(_internal=True)
    gen.generate_block(prog)
    code = gen.get_output()

    # Should have conditional structure (with unpacking, flag[0] becomes flag_0)
    assert "if flag[0]:" in code or "if flag_0:" in code
    # With unpacking, q[1] becomes q_1
    assert "quantum.measure(q[1])" in code or "quantum.measure(q_1)" in code
    # Should generate valid code
    assert "result(" in code


def test_ir_variable_renaming() -> None:
    """Test variable renaming to avoid conflicts."""
    prog = Main(
        result := QReg("result", 2),  # Conflicts with result() function
        array := CReg("array", 2),  # Conflicts with array() function
        Measure(result) > array,
    )
    gen = IRGuppyGenerator(_internal=True)
    gen.generate_block(prog)
    code = gen.get_output()

    # Should rename conflicting variables
    assert "result_reg" in code
    assert "array_reg" in code
    # Should use renamed variables correctly
    assert "quantum.measure_array" in code



# Copyright 2024 The PECOS Developers
# Licensed under the Apache License, Version 2.0

"""Generic transversal operations for CSS codes.

This module provides transversal CNOT generation that works with any CSS code
(surface codes, color codes, etc.) by abstracting over the code-specific details.

Transversal CNOT between two CSS code blocks applies CX(ctrl[i], tgt[i]) for all
data qubits i. This preserves the CSS structure: X errors on control propagate
to X errors on target, Z errors on target propagate to Z errors on control.
"""

import importlib.util
import sys
import tempfile
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Protocol, runtime_checkable


class CSSCodeType(Enum):
    """Supported CSS code types."""

    SURFACE = "surface"
    COLOR = "color"


@runtime_checkable
class CSSCodeSpec(Protocol):
    """Protocol for CSS code specifications."""

    @property
    def num_data(self) -> int:
        """Number of data qubits."""
        ...

    @property
    def num_x_stabilizers(self) -> int:
        """Number of X stabilizers."""
        ...

    @property
    def num_z_stabilizers(self) -> int:
        """Number of Z stabilizers."""
        ...

    def get_logical_x(self) -> tuple[int, ...]:
        """Qubits for logical X operator."""
        ...

    def get_logical_z(self) -> tuple[int, ...]:
        """Qubits for logical Z operator."""
        ...


@dataclass
class TransversalConfig:
    """Configuration for transversal CNOT generation."""

    code_type: CSSCodeType
    distance: int
    num_rounds: int = 1
    ctrl_logical_x: bool = False  # Apply logical X to control before CNOT


# Module state container (avoids global statement)
class _ModuleState:
    """Container for module-level mutable state."""

    temp_dir: Path | None = None
    css_transversal_cache: dict[str, dict] = {}  # noqa: RUF012


_state = _ModuleState()


def _get_temp_dir() -> Path:
    """Get or create temporary directory for generated code."""
    if _state.temp_dir is None:
        _state.temp_dir = Path(tempfile.mkdtemp(prefix="pecos_guppy_css_trans_"))
    return _state.temp_dir


def _get_surface_code_info(d: int) -> dict:
    """Get surface code parameters."""
    from pecos.qec.surface import SurfacePatch  # noqa: PLC0415

    patch = SurfacePatch.create(distance=d)
    geom = patch.geometry

    return {
        "num_data": d * d,
        "num_x_stab": len(geom.x_stabilizers),
        "num_z_stab": len(geom.z_stabilizers),
        "x_stabilizers": geom.x_stabilizers,
        "z_stabilizers": geom.z_stabilizers,
        "logical_x": [i * d for i in range(d)],  # Left column
        "logical_z": list(range(d)),  # Top row
        "struct_name": f"SurfaceCode_d{d}",
        "syndrome_name": f"SurfaceSyndrome_d{d}",
    }


def _get_color_code_info(d: int) -> dict:
    """Get color code parameters."""
    from pecos.qec.color import ColorCode488  # noqa: PLC0415

    code = ColorCode488.create(distance=d)

    return {
        "num_data": code.num_data,
        "num_x_stab": code.num_stabilizers,  # Same stabilizers for X and Z
        "num_z_stab": code.num_stabilizers,
        "stabilizers": code.stabilizers,
        "logical_x": list(code.get_logical_x()),
        "logical_z": list(code.get_logical_z()),
        "struct_name": f"ColorCode_d{d}",
        "syndrome_name": f"ColorSyndrome_d{d}",
    }


def generate_surface_transversal_source(d: int) -> str:
    """Generate transversal CNOT source for surface codes."""
    info = _get_surface_code_info(d)
    num_data = info["num_data"]
    num_x_stab = info["num_x_stab"]
    num_z_stab = info["num_z_stab"]

    lines = [
        f'"""Transversal CNOT for distance-{d} surface codes.',
        "",
        "Auto-generated for logical CNOT between two patches.",
        '"""',
        "",
        "from guppylang import guppy",
        "from guppylang.std.builtins import array, owned, result, comptime",
        "from guppylang.std.quantum import cx, discard, h, measure, measure_array, qubit, x",
        "from guppylang.std.qsystem import measure_and_reset",
        "",
        "",
    ]

    # Struct definitions
    lines.extend(
        [
            "@guppy.struct",
            f"class {info['struct_name']}:",
            f'    """Surface code patch d={d}."""',
            "",
            f"    data: array[qubit, {num_data}]",
            "",
            "",
            "@guppy.struct",
            f"class {info['syndrome_name']}:",
            f'    """Syndrome for d={d} surface code."""',
            "",
            f"    synx: array[bool, {num_x_stab}]",
            f"    synz: array[bool, {num_z_stab}]",
            "",
            "",
        ],
    )

    # X stabilizer measurements
    lines.append("# === X Stabilizer Measurements ===")
    lines.append("")

    for stab in info["x_stabilizers"]:
        lines.extend(
            [
                "@guppy",
                f"def measure_x_stab_{stab.index}(ax: qubit, data: array[qubit, {num_data}]) -> bool:",
                f'    """Measure X stabilizer {stab.index}."""',
                "    h(ax)",
            ],
        )
        lines.extend(f"    cx(ax, data[{q}])" for q in stab.data_qubits)
        lines.extend(
            [
                "    h(ax)",
                "    return measure_and_reset(ax)",
                "",
                "",
            ],
        )

    # Z stabilizer measurements
    lines.append("# === Z Stabilizer Measurements ===")
    lines.append("")

    for stab in info["z_stabilizers"]:
        lines.extend(
            [
                "@guppy",
                f"def measure_z_stab_{stab.index}(az: qubit, data: array[qubit, {num_data}]) -> bool:",
                f'    """Measure Z stabilizer {stab.index}."""',
            ],
        )
        lines.extend(f"    cx(data[{q}], az)" for q in stab.data_qubits)
        lines.extend(
            [
                "    return measure_and_reset(az)",
                "",
                "",
            ],
        )

    # Syndrome extraction
    x_calls = ", ".join(f"sx{s.index}" for s in info["x_stabilizers"])
    z_calls = ", ".join(f"sz{s.index}" for s in info["z_stabilizers"])

    lines.extend(
        [
            "# === Syndrome Extraction ===",
            "",
            "@guppy",
            "def syndrome_extraction(",
            f"    surf: {info['struct_name']},",
            "    ax: qubit,",
            "    az: qubit,",
            f") -> {info['syndrome_name']}:",
            '    """Extract full syndrome."""',
            "    # Z stabilizers",
        ],
    )

    lines.extend(f"    sz{stab.index} = measure_z_stab_{stab.index}(az, surf.data)" for stab in info["z_stabilizers"])

    lines.append("")
    lines.append("    # X stabilizers")

    lines.extend(f"    sx{stab.index} = measure_x_stab_{stab.index}(ax, surf.data)" for stab in info["x_stabilizers"])

    lines.extend(
        [
            "",
            f"    synx = array({x_calls})",
            f"    synz = array({z_calls})",
            "",
            f"    return {info['syndrome_name']}(synx, synz)",
            "",
            "",
        ],
    )

    # Initialization
    lines.extend(
        [
            "# === Initialization ===",
            "",
            "@guppy",
            f"def init_z_basis(surf: {info['struct_name']}, ax: qubit) -> array[bool, {num_x_stab}]:",
            '    """Initialize logical |0_L> and extract initial X syndrome."""',
        ],
    )

    lines.extend(f"    sx{stab.index} = measure_x_stab_{stab.index}(ax, surf.data)" for stab in info["x_stabilizers"])

    lines.extend(
        [
            "",
            f"    return array({x_calls})",
            "",
            "",
        ],
    )

    # Apply logical X
    lines.extend(
        [
            "@guppy",
            f"def apply_logical_x(surf: {info['struct_name']}) -> None:",
            '    """Apply logical X operator."""',
        ],
    )
    lines.extend(f"    x(surf.data[{q}])" for q in info["logical_x"])
    lines.extend(
        [
            "",
            "",
        ],
    )

    # Transversal CNOT
    lines.extend(
        [
            "# === Transversal Operations ===",
            "",
            "@guppy",
            f"def transversal_cnot(ctrl: {info['struct_name']}, tgt: {info['struct_name']}) -> None:",
            '    """Apply transversal CNOT: ctrl[i] controls tgt[i]."""',
            f"    for i in range({num_data}):",
            "        cx(ctrl.data[i], tgt.data[i])",
            "",
            "",
        ],
    )

    # Measurement
    lines.extend(
        [
            "# === Measurement ===",
            "",
            "@guppy",
            f"def measure_z_basis(surf: {info['struct_name']} @ owned) -> array[bool, {num_data}]:",
            '    """Destructively measure in Z basis."""',
            "    return measure_array(surf.data)",
            "",
            "",
        ],
    )

    # Factory functions
    _add_transversal_factory_functions(lines, info, num_data, num_x_stab)

    return "\n".join(lines)


def generate_color_transversal_source(d: int) -> str:
    """Generate transversal CNOT source for color codes."""
    info = _get_color_code_info(d)
    num_data = info["num_data"]
    num_stab = info["num_x_stab"]  # Same for X and Z

    lines = [
        f'"""Transversal CNOT for distance-{d} color codes.',
        "",
        "Auto-generated for logical CNOT between two patches.",
        '"""',
        "",
        "from guppylang import guppy",
        "from guppylang.std.builtins import array, owned, result, comptime",
        "from guppylang.std.quantum import cx, discard, h, measure, measure_array, qubit, x",
        "from guppylang.std.qsystem import measure_and_reset",
        "",
        "",
    ]

    # Struct definitions
    lines.extend(
        [
            "@guppy.struct",
            f"class {info['struct_name']}:",
            f'    """Color code patch d={d}."""',
            "",
            f"    data: array[qubit, {num_data}]",
            "",
            "",
            "@guppy.struct",
            f"class {info['syndrome_name']}:",
            f'    """Syndrome for d={d} color code."""',
            "",
            f"    synx: array[bool, {num_stab}]",
            f"    synz: array[bool, {num_stab}]",
            "",
            "",
        ],
    )

    # X stabilizer measurements (H-CNOT-H pattern)
    lines.append("# === X Stabilizer Measurements ===")
    lines.append("")

    for stab in info["stabilizers"]:
        lines.extend(
            [
                "@guppy",
                f"def measure_x_stab_{stab.index}(ax: qubit, data: array[qubit, {num_data}]) -> bool:",
                f'    """Measure X stabilizer {stab.index} ({stab.color})."""',
                "    h(ax)",
            ],
        )
        lines.extend(f"    cx(ax, data[{q}])" for q in stab.qubits)
        lines.extend(
            [
                "    h(ax)",
                "    return measure_and_reset(ax)",
                "",
                "",
            ],
        )

    # Z stabilizer measurements
    lines.append("# === Z Stabilizer Measurements ===")
    lines.append("")

    for stab in info["stabilizers"]:
        lines.extend(
            [
                "@guppy",
                f"def measure_z_stab_{stab.index}(az: qubit, data: array[qubit, {num_data}]) -> bool:",
                f'    """Measure Z stabilizer {stab.index} ({stab.color})."""',
            ],
        )
        lines.extend(f"    cx(data[{q}], az)" for q in stab.qubits)
        lines.extend(
            [
                "    return measure_and_reset(az)",
                "",
                "",
            ],
        )

    # Syndrome extraction
    x_calls = ", ".join(f"sx{s.index}" for s in info["stabilizers"])
    z_calls = ", ".join(f"sz{s.index}" for s in info["stabilizers"])

    lines.extend(
        [
            "# === Syndrome Extraction ===",
            "",
            "@guppy",
            "def syndrome_extraction(",
            f"    code: {info['struct_name']},",
            "    ax: qubit,",
            "    az: qubit,",
            f") -> {info['syndrome_name']}:",
            '    """Extract full syndrome."""',
            "    # Z stabilizers first",
        ],
    )

    lines.extend(f"    sz{stab.index} = measure_z_stab_{stab.index}(az, code.data)" for stab in info["stabilizers"])

    lines.append("")
    lines.append("    # X stabilizers")

    lines.extend(f"    sx{stab.index} = measure_x_stab_{stab.index}(ax, code.data)" for stab in info["stabilizers"])

    lines.extend(
        [
            "",
            f"    synx = array({x_calls})",
            f"    synz = array({z_calls})",
            "",
            f"    return {info['syndrome_name']}(synx, synz)",
            "",
            "",
        ],
    )

    # Initialization
    lines.extend(
        [
            "# === Initialization ===",
            "",
            "@guppy",
            f"def init_z_basis(code: {info['struct_name']}, ax: qubit) -> array[bool, {num_stab}]:",
            '    """Initialize logical |0_L> and extract initial X syndrome."""',
        ],
    )

    lines.extend(f"    sx{stab.index} = measure_x_stab_{stab.index}(ax, code.data)" for stab in info["stabilizers"])

    lines.extend(
        [
            "",
            f"    return array({x_calls})",
            "",
            "",
        ],
    )

    # Apply logical X
    lines.extend(
        [
            "@guppy",
            f"def apply_logical_x(code: {info['struct_name']}) -> None:",
            '    """Apply logical X operator."""',
        ],
    )
    lines.extend(f"    x(code.data[{q}])" for q in info["logical_x"])
    lines.extend(
        [
            "",
            "",
        ],
    )

    # Transversal CNOT
    lines.extend(
        [
            "# === Transversal Operations ===",
            "",
            "@guppy",
            f"def transversal_cnot(ctrl: {info['struct_name']}, tgt: {info['struct_name']}) -> None:",
            '    """Apply transversal CNOT: ctrl[i] controls tgt[i]."""',
            f"    for i in range({num_data}):",
            "        cx(ctrl.data[i], tgt.data[i])",
            "",
            "",
        ],
    )

    # Measurement
    lines.extend(
        [
            "# === Measurement ===",
            "",
            "@guppy",
            f"def measure_z_basis(code: {info['struct_name']} @ owned) -> array[bool, {num_data}]:",
            '    """Destructively measure in Z basis."""',
            "    return measure_array(code.data)",
            "",
            "",
        ],
    )

    # Factory functions
    _add_color_transversal_factory_functions(lines, info, num_data, num_stab)

    return "\n".join(lines)


def _add_transversal_factory_functions(
    lines: list[str],
    info: dict,
    num_data: int,
    _num_stab: int,
) -> None:
    """Add factory functions for surface code transversal experiments."""
    struct_name = info["struct_name"]

    lines.extend(
        [
            "# === Transversal Experiments ===",
            "",
            "def make_transversal_cnot(num_rounds: int):",
            '    """Create transversal CNOT experiment: |0_L>|0_L> -> |0_L>|0_L>."""',
            "",
            "    @guppy",
            "    def transversal_cnot_exp() -> None:",
            '        """Transversal CNOT experiment."""',
            f"        ctrl_data = array(qubit() for _ in range({num_data}))",
            f"        tgt_data = array(qubit() for _ in range({num_data}))",
            "        ax_ctrl = qubit()",
            "        az_ctrl = qubit()",
            "        ax_tgt = qubit()",
            "        az_tgt = qubit()",
            "",
            f"        ctrl = {struct_name}(ctrl_data)",
            f"        tgt = {struct_name}(tgt_data)",
            "",
            "        # Initialize both patches in |0_L>",
            "        init_syn_ctrl = init_z_basis(ctrl, ax_ctrl)",
            "        init_syn_tgt = init_z_basis(tgt, ax_tgt)",
            '        result("init_synx_ctrl", init_syn_ctrl)',
            '        result("init_synx_tgt", init_syn_tgt)',
            "",
            "        # Apply transversal CNOT",
            "        transversal_cnot(ctrl, tgt)",
            "",
            "        # Syndrome extraction rounds",
            "        for _t in range(comptime(num_rounds)):",
            "            syn_ctrl = syndrome_extraction(ctrl, ax_ctrl, az_ctrl)",
            "            syn_tgt = syndrome_extraction(tgt, ax_tgt, az_tgt)",
            '            result("synx_ctrl", syn_ctrl.synx)',
            '            result("synz_ctrl", syn_ctrl.synz)',
            '            result("synx_tgt", syn_tgt.synx)',
            '            result("synz_tgt", syn_tgt.synz)',
            "",
            "        # Final measurement",
            "        final_ctrl = measure_z_basis(ctrl)",
            "        final_tgt = measure_z_basis(tgt)",
            '        result("final_ctrl", final_ctrl)',
            '        result("final_tgt", final_tgt)',
            "",
            "        discard(ax_ctrl)",
            "        discard(az_ctrl)",
            "        discard(ax_tgt)",
            "        discard(az_tgt)",
            "",
            "    return transversal_cnot_exp",
            "",
            "",
            "def make_transversal_cnot_with_x(num_rounds: int):",
            '    """Create transversal CNOT experiment: |1_L>|0_L> -> |1_L>|1_L>."""',
            "",
            "    @guppy",
            "    def transversal_cnot_with_x_exp() -> None:",
            '        """Transversal CNOT with X experiment."""',
            f"        ctrl_data = array(qubit() for _ in range({num_data}))",
            f"        tgt_data = array(qubit() for _ in range({num_data}))",
            "        ax_ctrl = qubit()",
            "        az_ctrl = qubit()",
            "        ax_tgt = qubit()",
            "        az_tgt = qubit()",
            "",
            f"        ctrl = {struct_name}(ctrl_data)",
            f"        tgt = {struct_name}(tgt_data)",
            "",
            "        # Initialize both patches in |0_L>",
            "        init_syn_ctrl = init_z_basis(ctrl, ax_ctrl)",
            "        init_syn_tgt = init_z_basis(tgt, ax_tgt)",
            '        result("init_synx_ctrl", init_syn_ctrl)',
            '        result("init_synx_tgt", init_syn_tgt)',
            "",
            "        # Apply logical X to control to get |1_L>",
            "        apply_logical_x(ctrl)",
            "",
            "        # Apply transversal CNOT",
            "        transversal_cnot(ctrl, tgt)",
            "",
            "        # Syndrome extraction rounds",
            "        for _t in range(comptime(num_rounds)):",
            "            syn_ctrl = syndrome_extraction(ctrl, ax_ctrl, az_ctrl)",
            "            syn_tgt = syndrome_extraction(tgt, ax_tgt, az_tgt)",
            '            result("synx_ctrl", syn_ctrl.synx)',
            '            result("synz_ctrl", syn_ctrl.synz)',
            '            result("synx_tgt", syn_tgt.synx)',
            '            result("synz_tgt", syn_tgt.synz)',
            "",
            "        # Final measurement",
            "        final_ctrl = measure_z_basis(ctrl)",
            "        final_tgt = measure_z_basis(tgt)",
            '        result("final_ctrl", final_ctrl)',
            '        result("final_tgt", final_tgt)',
            "",
            "        discard(ax_ctrl)",
            "        discard(az_ctrl)",
            "        discard(ax_tgt)",
            "        discard(az_tgt)",
            "",
            "    return transversal_cnot_with_x_exp",
            "",
        ],
    )


def _add_color_transversal_factory_functions(
    lines: list[str],
    info: dict,
    num_data: int,
    _num_stab: int,
) -> None:
    """Add factory functions for color code transversal experiments."""
    struct_name = info["struct_name"]

    lines.extend(
        [
            "# === Transversal Experiments ===",
            "",
            "def make_transversal_cnot(num_rounds: int):",
            '    """Create transversal CNOT experiment: |0_L>|0_L> -> |0_L>|0_L>."""',
            "",
            "    @guppy",
            "    def transversal_cnot_exp() -> None:",
            '        """Transversal CNOT experiment."""',
            f"        ctrl_data = array(qubit() for _ in range({num_data}))",
            f"        tgt_data = array(qubit() for _ in range({num_data}))",
            "        ax_ctrl = qubit()",
            "        az_ctrl = qubit()",
            "        ax_tgt = qubit()",
            "        az_tgt = qubit()",
            "",
            f"        ctrl = {struct_name}(ctrl_data)",
            f"        tgt = {struct_name}(tgt_data)",
            "",
            "        # Initialize both patches in |0_L>",
            "        init_syn_ctrl = init_z_basis(ctrl, ax_ctrl)",
            "        init_syn_tgt = init_z_basis(tgt, ax_tgt)",
            '        result("init_synx_ctrl", init_syn_ctrl)',
            '        result("init_synx_tgt", init_syn_tgt)',
            "",
            "        # Apply transversal CNOT",
            "        transversal_cnot(ctrl, tgt)",
            "",
            "        # Syndrome extraction rounds",
            "        for _t in range(comptime(num_rounds)):",
            "            syn_ctrl = syndrome_extraction(ctrl, ax_ctrl, az_ctrl)",
            "            syn_tgt = syndrome_extraction(tgt, ax_tgt, az_tgt)",
            '            result("synx_ctrl", syn_ctrl.synx)',
            '            result("synz_ctrl", syn_ctrl.synz)',
            '            result("synx_tgt", syn_tgt.synx)',
            '            result("synz_tgt", syn_tgt.synz)',
            "",
            "        # Final measurement",
            "        final_ctrl = measure_z_basis(ctrl)",
            "        final_tgt = measure_z_basis(tgt)",
            '        result("final_ctrl", final_ctrl)',
            '        result("final_tgt", final_tgt)',
            "",
            "        discard(ax_ctrl)",
            "        discard(az_ctrl)",
            "        discard(ax_tgt)",
            "        discard(az_tgt)",
            "",
            "    return transversal_cnot_exp",
            "",
            "",
            "def make_transversal_cnot_with_x(num_rounds: int):",
            '    """Create transversal CNOT experiment: |1_L>|0_L> -> |1_L>|1_L>."""',
            "",
            "    @guppy",
            "    def transversal_cnot_with_x_exp() -> None:",
            '        """Transversal CNOT with X experiment."""',
            f"        ctrl_data = array(qubit() for _ in range({num_data}))",
            f"        tgt_data = array(qubit() for _ in range({num_data}))",
            "        ax_ctrl = qubit()",
            "        az_ctrl = qubit()",
            "        ax_tgt = qubit()",
            "        az_tgt = qubit()",
            "",
            f"        ctrl = {struct_name}(ctrl_data)",
            f"        tgt = {struct_name}(tgt_data)",
            "",
            "        # Initialize both patches in |0_L>",
            "        init_syn_ctrl = init_z_basis(ctrl, ax_ctrl)",
            "        init_syn_tgt = init_z_basis(tgt, ax_tgt)",
            '        result("init_synx_ctrl", init_syn_ctrl)',
            '        result("init_synx_tgt", init_syn_tgt)',
            "",
            "        # Apply logical X to control to get |1_L>",
            "        apply_logical_x(ctrl)",
            "",
            "        # Apply transversal CNOT",
            "        transversal_cnot(ctrl, tgt)",
            "",
            "        # Syndrome extraction rounds",
            "        for _t in range(comptime(num_rounds)):",
            "            syn_ctrl = syndrome_extraction(ctrl, ax_ctrl, az_ctrl)",
            "            syn_tgt = syndrome_extraction(tgt, ax_tgt, az_tgt)",
            '            result("synx_ctrl", syn_ctrl.synx)',
            '            result("synz_ctrl", syn_ctrl.synz)',
            '            result("synx_tgt", syn_tgt.synx)',
            '            result("synz_tgt", syn_tgt.synz)',
            "",
            "        # Final measurement",
            "        final_ctrl = measure_z_basis(ctrl)",
            "        final_tgt = measure_z_basis(tgt)",
            '        result("final_ctrl", final_ctrl)',
            '        result("final_tgt", final_tgt)',
            "",
            "        discard(ax_ctrl)",
            "        discard(az_ctrl)",
            "        discard(ax_tgt)",
            "        discard(az_tgt)",
            "",
            "    return transversal_cnot_with_x_exp",
            "",
        ],
    )


def _load_css_transversal_module(code_type: CSSCodeType, d: int) -> dict:
    """Load a transversal module for the given code type and distance."""
    cache_key = f"{code_type.value}_d{d}"

    if cache_key in _state.css_transversal_cache:
        return _state.css_transversal_cache[cache_key]

    # Generate source based on code type
    if code_type == CSSCodeType.SURFACE:
        source = generate_surface_transversal_source(d)
    elif code_type == CSSCodeType.COLOR:
        source = generate_color_transversal_source(d)
    else:
        msg = f"Unsupported code type: {code_type}"
        raise ValueError(msg)

    # Write to temp file
    temp_dir = _get_temp_dir()
    temp_file = temp_dir / f"css_trans_{cache_key}.py"
    temp_file.write_text(source)

    # Load module
    module_name = f"pecos._generated.css_trans_{cache_key}"
    spec = importlib.util.spec_from_file_location(module_name, temp_file)
    if spec is None or spec.loader is None:
        msg = f"Failed to create module spec for {temp_file}"
        raise RuntimeError(msg)

    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)

    _state.css_transversal_cache[cache_key] = vars(module)
    return _state.css_transversal_cache[cache_key]


# === Public API ===


def make_css_transversal_cnot(
    code_type: CSSCodeType | str,
    distance: int,
    num_rounds: int = 1,
) -> object:
    """Create a transversal CNOT experiment for any CSS code.

    Args:
        code_type: Type of CSS code ("surface" or "color")
        distance: Code distance (must be odd >= 3)
        num_rounds: Number of syndrome extraction rounds

    Returns:
        Guppy function for the experiment

    Example:
        >>> prog = make_css_transversal_cnot("color", distance=3, num_rounds=1)
        >>> result = prog.emulator(n_qubits=18).stabilizer_sim().run()
    """
    if isinstance(code_type, str):
        code_type = CSSCodeType(code_type)

    if distance < 3 or distance % 2 == 0:
        msg = f"Distance must be odd >= 3, got {distance}"
        raise ValueError(msg)

    module = _load_css_transversal_module(code_type, distance)
    return module["make_transversal_cnot"](num_rounds)


def make_css_transversal_cnot_with_x(
    code_type: CSSCodeType | str,
    distance: int,
    num_rounds: int = 1,
) -> object:
    """Create a transversal CNOT experiment with logical X on control.

    This tests |1_L>|0_L> -> |1_L>|1_L>.

    Args:
        code_type: Type of CSS code ("surface" or "color")
        distance: Code distance (must be odd >= 3)
        num_rounds: Number of syndrome extraction rounds

    Returns:
        Guppy function for the experiment
    """
    if isinstance(code_type, str):
        code_type = CSSCodeType(code_type)

    if distance < 3 or distance % 2 == 0:
        msg = f"Distance must be odd >= 3, got {distance}"
        raise ValueError(msg)

    module = _load_css_transversal_module(code_type, distance)
    return module["make_transversal_cnot_with_x"](num_rounds)


def get_transversal_num_qubits(code_type: CSSCodeType | str, distance: int) -> int:
    """Get total qubit count for transversal CNOT between two patches.

    Args:
        code_type: Type of CSS code
        distance: Code distance

    Returns:
        Total qubits needed (2 * num_data + 4 ancillas)
    """
    if isinstance(code_type, str):
        code_type = CSSCodeType(code_type)

    if code_type == CSSCodeType.SURFACE:
        num_data = distance * distance
    elif code_type == CSSCodeType.COLOR:
        from pecos.qec.color import ColorCode488  # noqa: PLC0415

        code = ColorCode488.create(distance=distance)
        num_data = code.num_data
    else:
        msg = f"Unsupported code type: {code_type}"
        raise ValueError(msg)

    # Two patches + 4 ancillas (ax, az for each patch)
    return 2 * num_data + 4


# === Convenience functions for specific codes ===


def make_color_transversal_cnot(distance: int, num_rounds: int = 1) -> object:
    """Create transversal CNOT for color codes."""
    return make_css_transversal_cnot(CSSCodeType.COLOR, distance, num_rounds)


def make_color_transversal_cnot_with_x(distance: int, num_rounds: int = 1) -> object:
    """Create transversal CNOT with X for color codes."""
    return make_css_transversal_cnot_with_x(CSSCodeType.COLOR, distance, num_rounds)


def make_color_transversal_cnot_d3(num_rounds: int = 1) -> object:
    """Create d=3 transversal CNOT for color codes."""
    return make_color_transversal_cnot(3, num_rounds)


def make_color_transversal_cnot_with_x_d3(num_rounds: int = 1) -> object:
    """Create d=3 transversal CNOT with X for color codes."""
    return make_color_transversal_cnot_with_x(3, num_rounds)


def make_surface_transversal_cnot(distance: int, num_rounds: int = 1) -> object:
    """Create transversal CNOT for surface codes."""
    return make_css_transversal_cnot(CSSCodeType.SURFACE, distance, num_rounds)


def make_surface_transversal_cnot_with_x(distance: int, num_rounds: int = 1) -> object:
    """Create transversal CNOT with X for surface codes."""
    return make_css_transversal_cnot_with_x(CSSCodeType.SURFACE, distance, num_rounds)



"""Lookup table decoder for the Steane 7-qubit code.

This module provides lookup table decoding implementations for the Steane 7-qubit quantum error correction code,
enabling syndrome-based error correction through pre-computed correction tables.
"""

# Copyright 2024 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

from pecos.slr import Block, Comment, If
from pecos.slr.qeclib import qubit

if TYPE_CHECKING:
    from pecos.slr import Bit, CReg, QReg


class FlagLookupQASM(Block):
    """Flag-based lookup decoder for Steane code syndrome decoding.

    This class implements a lookup table-based decoder that uses flag qubits
    to determine the appropriate error correction based on measured syndromes.
    """

    def __init__(
        self,
        basis: str,
        syn: CReg,
        syndromes: CReg,
        raw_syn: CReg,
        pf: Bit,
        flag: CReg,
        flags: CReg,
        scratch: CReg,
    ) -> None:
        """Initialize flag-based lookup decoder for Steane code.

        Args:
            basis: The basis ('X' or 'Z') for syndrome decoding.
            syn: Classical register containing current syndrome values.
            syndromes: Classical register for processed syndrome values.
            raw_syn: Classical register storing raw syndrome values.
            pf: Pauli frame bit for tracking logical operations.
            flag: Classical register containing flag qubit measurements.
            flags: Classical register for flag status.
            scratch: Classical register used for intermediate calculations.
        """
        super().__init__()

        # qasm_syn_decoder('X', syn_x, flag_x, 'last_raw_syn_x', 'pf_z1')
        # qasm_syn_decoder(basis_check, syn, flag, raw_syn, pf, pf_index=0)

        self.extend(
            Comment(
                f"""
=========================
BEGIN Run {basis} decoder
=========================\n""",
            ),
            If(flags != 0).Then(syndromes.set(syn ^ raw_syn)),
            If(flags == 0).Then(syndromes.set(0)),
            Comment("\napply corrections"),
            If(syndromes == 2).Then(pf.set(pf ^ 1)),
            If(syndromes == 4).Then(pf.set(pf ^ 1)),
            If(syndromes == 6).Then(pf.set(pf ^ 1)),
            Comment(),
            Comment("alter correction based on flags"),
            Comment("==============================="),
            Comment(),
            Comment("1&2 (1 -> 2)"),
            Comment("------------"),
            scratch.set(0),
            If(flag == 1).Then(scratch[0].set(1)),
            If(syndromes == 2).Then(scratch[1].set(1)),
            Comment(),
            scratch[2].set(scratch[0] & scratch[1]),
            If(scratch[2] == 1).Then(pf.set(pf ^ 1)),
            Comment(),
            Comment("1&4 (1 -> 3)"),
            Comment("------------"),
            scratch.set(0),
            If(flag == 1).Then(scratch[0].set(1)),
            If(syndromes == 4).Then(scratch[1].set(1)),
            Comment(),
            scratch[2].set(scratch[0] & scratch[1]),
            If(scratch[2] == 1).Then(pf.set(pf ^ 1)),
            Comment(),
            Comment("6&4 (2,3 -> 3)"),
            Comment("------------"),
            scratch.set(0),
            If(flag == 6).Then(scratch[0].set(1)),
            If(syndromes == 4).Then(scratch[1].set(1)),
            Comment(),
            scratch[2].set(scratch[0] & scratch[1]),
            If(scratch[2] == 1).Then(pf.set(pf ^ 1)),
            Comment(),
            If(flags != 0).Then(raw_syn.set(syn)),
            Comment(),
            Comment("========================="),
            Comment(f"END Run {basis} decoder"),
            Comment("=========================\n"),
        )


class FlagLookupQASMActiveCorrectionX(Block):
    """Active X-basis correction using flag-based lookup for Steane code.

    This class applies X-basis error corrections based on syndrome measurements
    and flag qubit information using a lookup table approach.
    """

    def __init__(
        self,
        qubits: QReg,
        syn: CReg,
        syndromes: CReg,
        raw_syn: CReg,
        pf: Bit,
        flag: CReg,
        flags: CReg,
        scratch: CReg,
        pf_bit_copy: Bit | None = None,
    ) -> None:
        """Initialize X-basis active error correction with flag-based lookup.

        Args:
            qubits: Quantum register containing the logical qubit data.
            syn: Classical register containing current syndrome values.
            syndromes: Classical register for processed syndrome values.
            raw_syn: Classical register storing raw syndrome values.
            pf: Pauli frame bit for tracking logical operations.
            flag: Classical register containing flag qubit measurements.
            flags: Classical register for flag status.
            scratch: Classical register used for intermediate calculations.
            pf_bit_copy: Optional bit to copy the Pauli frame value to.
        """
        super().__init__()
        # qasm_syn_decoder('X', syn_x, flag_x, 'last_raw_syn_x', 'pf_z1')
        # qasm_syn_decoder(basis_check, syn, flag, raw_syn, pf, pf_index=0)
        q = qubits

        self.extend(
            FlagLookupQASM(
                basis="X",
                syn=syn,
                syndromes=syndromes,
                raw_syn=raw_syn,
                pf=pf,
                flag=flag,
                flags=flags,
                scratch=scratch,
            ),
        )

        if pf_bit_copy is not None:
            self.extend(
                Comment(),
                Comment("copy Pauli frame"),
                pf_bit_copy.set(pf),
            )

        self.extend(
            Comment(),
            Comment(),
            Comment("ACTIVE ERROR CORRECTION FOR X SYNDROMES"),
            Comment(),
            scratch.set(0),
            Comment(),
            Comment("only part that differs for X vs Z syns V"),
            If(syndromes[0] == 1).Then(scratch.set(scratch ^ 1)),
            If(syndromes[1] == 1).Then(scratch.set(scratch ^ 12)),
            If(syndromes[2] == 1).Then(scratch.set(scratch ^ 48)),
            Comment(),
            Comment("logical operator"),
            If(pf == 1).Then(scratch.set(scratch ^ 112)),
            Comment(),
            If(scratch[0] == 1).Then(qubit.Z(q[0])),
            Comment("not possible for X stabilizers V"),
            Comment(
                f"if({scratch[1]} == 1) z {q[1]};",
            ),  # If(scratch[2] == 1).Then(qubit.Z(q[1])),
            If(scratch[2] == 1).Then(qubit.Z(q[2])),
            If(scratch[3] == 1).Then(qubit.Z(q[3])),
            If(scratch[4] == 1).Then(qubit.Z(q[4])),
            If(scratch[5] == 1).Then(qubit.Z(q[5])),
            If(scratch[6] == 1).Then(qubit.Z(q[6])),
            Comment(),
            pf.set(0),
            Comment(f"{syndromes} = 0;"),
            raw_syn.set(0),
            Comment(f"{syn} = 0;"),
            Comment(f"{flag} = 0;"),
            Comment(f"{flags} = 0;"),
            Comment(),
            Comment(),
        )


class FlagLookupQASMActiveCorrectionZ(Block):
    """Active Z-basis correction using flag-based lookup for Steane code.

    This class applies Z-basis error corrections based on syndrome measurements
    and flag qubit information using a lookup table approach.
    """

    def __init__(
        self,
        qubits: QReg,
        syn: CReg,
        syndromes: CReg,
        raw_syn: CReg,
        pf: Bit,
        flag: CReg,
        flags: CReg,
        scratch: CReg,
        pf_bit_copy: Bit = None,
    ) -> None:
        """Initialize Z-basis active error correction with flag-based lookup.

        Args:
            qubits: Quantum register containing the logical qubit data.
            syn: Classical register containing current syndrome values.
            syndromes: Classical register for processed syndrome values.
            raw_syn: Classical register storing raw syndrome values.
            pf: Pauli frame bit for tracking logical operations.
            flag: Classical register containing flag qubit measurements.
            flags: Classical register for flag status.
            scratch: Classical register used for intermediate calculations.
            pf_bit_copy: Optional bit to copy the Pauli frame value to.
        """
        super().__init__()
        q = qubits

        # qasm_syn_decoder('X', syn_x, flag_x, 'last_raw_syn_x', 'pf_z1')
        # qasm_syn_decoder(basis_check, syn, flag, raw_syn, pf, pf_index=0)
        self.extend(
            FlagLookupQASM(
                basis="Z",
                syn=syn,
                syndromes=syndromes,
                raw_syn=raw_syn,
                pf=pf,
                flag=flag,
                flags=flags,
                scratch=scratch,
            ),
        )

        if pf_bit_copy is not None:
            self.extend(
                Comment(),
                Comment("copy Pauli frame"),
                pf_bit_copy.set(pf),
            )

        self.extend(
            Comment(),
            Comment(),
            Comment("ACTIVE ERROR CORRECTION FOR Z SYNDROMES"),
            Comment(),
            scratch.set(0),
            Comment(),
            Comment("only part that differs for X vs Z syns V"),
            If(syndromes[0] == 1).Then(scratch.set(scratch ^ 14)),
            If(syndromes[1] == 1).Then(scratch.set(scratch ^ 12)),
            If(syndromes[2] == 1).Then(scratch.set(scratch ^ 48)),
            Comment(),
            Comment("logical operator"),
            If(pf == 1).Then(scratch.set(scratch ^ 112)),
            Comment(),
            Comment("not possible for X stabilizers V"),
            Comment(f"if({scratch[0]} == 1) z {q[0]};"),
            If(scratch[1] == 1).Then(qubit.X(q[1])),
            If(scratch[2] == 1).Then(qubit.X(q[2])),
            If(scratch[3] == 1).Then(qubit.X(q[3])),
            If(scratch[4] == 1).Then(qubit.X(q[4])),
            If(scratch[5] == 1).Then(qubit.X(q[5])),
            If(scratch[6] == 1).Then(qubit.X(q[6])),
            Comment(),
            pf.set(0),
            Comment(f"{syndromes} = 0;"),
            raw_syn.set(0),
            Comment(f"{syn} = 0;"),
            Comment(f"{flag} = 0;"),
            Comment(f"{flags} = 0;"),
            Comment(),
        )



"""Data flow analysis for SLR to Guppy code generation.

This module provides data flow analysis to track how quantum and classical values
flow through a program, particularly tracking measurement results and their usage.

The key insight is that we need to distinguish between:
1. Operations BEFORE measurement (don't require unpacking)
2. Operations AFTER measurement that use the SAME qubit (require unpacking for replacement)
3. Operations AFTER measurement that use DIFFERENT qubits (don't require unpacking)

Current heuristics over-approximate by treating ANY operation after ANY measurement
as requiring unpacking, leading to unnecessary array unpacking.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from pecos.slr import Block as SLRBlock


@dataclass
class ValueUse:
    """Represents a use of a value (qubit or classical bit)."""

    array_name: str
    index: int
    position: int  # Position in operation sequence
    operation_type: str  # e.g., "gate", "measurement", "condition"
    is_consuming: bool = False  # True if this use consumes the value (e.g., measurement)


@dataclass
class DataFlowInfo:
    """Information about data flow for a single array element."""

    array_name: str
    index: int
    is_classical: bool

    # Track all uses of this element
    uses: list[ValueUse] = field(default_factory=list)

    # Track consumption points (measurements)
    consumed_at: list[int] = field(default_factory=list)

    # Track replacements (e.g., Prep after measurement)
    replaced_at: list[int] = field(default_factory=list)

    def add_use(
        self,
        position: int,
        operation_type: str,
        *,
        is_consuming: bool = False,
    ) -> None:
        """Add a use of this value."""
        use = ValueUse(
            array_name=self.array_name,
            index=self.index,
            position=position,
            operation_type=operation_type,
            is_consuming=is_consuming,
        )
        self.uses.append(use)

        if is_consuming:
            self.consumed_at.append(position)

    def add_replacement(self, position: int) -> None:
        """Mark that this value is replaced at a position (e.g., Prep)."""
        self.replaced_at.append(position)

    def has_use_after_consumption(self) -> bool:
        """Check if this element is used after being consumed.

        This is the key analysis for determining if unpacking is needed.
        If a qubit is measured and then used again (not just replaced),
        we need unpacking to handle the replacement properly.
        """
        if not self.consumed_at:
            return False

        # Find the first consumption point
        first_consumption = min(self.consumed_at)

        # Check if there are any non-replacement uses after consumption
        for use in self.uses:
            if use.position > first_consumption and use.position not in self.replaced_at:
                # This is a real use after consumption, not just replacement
                # However, we need to check if it's AFTER replacement
                # Find if there's a replacement between consumption and this use
                replacements_between = [r for r in self.replaced_at if first_consumption < r < use.position]

                if not replacements_between:
                    # Use after consumption with no replacement in between
                    # This requires unpacking
                    return True

        return False

    def requires_unpacking_for_flow(self) -> bool:
        """Determine if this element requires unpacking based on data flow.

        This is more precise than the heuristic approach:
        - Classical values can be used multiple times without issue
        - Quantum values can be used multiple times if not measured
        - Quantum values that are measured and then used require unpacking
        """
        if self.is_classical:
            # Classical values can be read multiple times
            return False

        # Quantum values: check if used after consumption
        return self.has_use_after_consumption()


@dataclass
class DataFlowAnalysis:
    """Complete data flow analysis for a block."""

    # Map from (array_name, index) to DataFlowInfo
    element_flows: dict[tuple[str, int], DataFlowInfo] = field(default_factory=dict)

    # Track conditionally accessed elements
    conditional_accesses: set[tuple[str, int]] = field(default_factory=set)

    def get_or_create_flow(
        self,
        array_name: str,
        index: int,
        is_classical: bool,
    ) -> DataFlowInfo:
        """Get or create data flow info for an array element."""
        key = (array_name, index)
        if key not in self.element_flows:
            self.element_flows[key] = DataFlowInfo(
                array_name=array_name,
                index=index,
                is_classical=is_classical,
            )
        return self.element_flows[key]

    def add_gate_use(self, array_name: str, index: int, position: int) -> None:
        """Record a gate operation on an array element."""
        flow = self.get_or_create_flow(array_name, index, is_classical=False)
        flow.add_use(position, "gate", is_consuming=False)

    def add_measurement(
        self,
        quantum_array: str,
        quantum_index: int,
        position: int,
        classical_array: str | None = None,
        classical_index: int | None = None,
    ) -> None:
        """Record a measurement operation."""
        # Quantum side: consumption
        q_flow = self.get_or_create_flow(
            quantum_array,
            quantum_index,
            is_classical=False,
        )
        q_flow.add_use(position, "measurement", is_consuming=True)

        # Classical side: creation (if specified)
        if classical_array is not None and classical_index is not None:
            c_flow = self.get_or_create_flow(
                classical_array,
                classical_index,
                is_classical=True,
            )
            c_flow.add_use(position, "measurement_result", is_consuming=False)

    def add_preparation(self, array_name: str, index: int, position: int) -> None:
        """Record a preparation/reset operation (replaces a qubit)."""
        flow = self.get_or_create_flow(array_name, index, is_classical=False)
        flow.add_use(position, "preparation", is_consuming=False)
        flow.add_replacement(position)

    def add_conditional_use(
        self,
        array_name: str,
        index: int,
        position: int,
        is_classical: bool,
    ) -> None:
        """Record a conditional use of an array element."""
        flow = self.get_or_create_flow(array_name, index, is_classical)
        flow.add_use(position, "condition", is_consuming=False)
        self.conditional_accesses.add((array_name, index))

    def elements_requiring_unpacking(self) -> set[tuple[str, int]]:
        """Get the set of array elements that require unpacking based on data flow."""
        requiring_unpacking = set()

        for key, flow in self.element_flows.items():
            if flow.requires_unpacking_for_flow():
                requiring_unpacking.add(key)

        return requiring_unpacking

    def array_requires_unpacking(self, array_name: str) -> bool:
        """Check if an entire array requires unpacking based on data flow."""
        for key, flow in self.element_flows.items():
            if key[0] == array_name and flow.requires_unpacking_for_flow():
                return True
        return False


class DataFlowAnalyzer:
    """Analyzes data flow in SLR blocks."""

    def __init__(self):
        self.position_counter = 0
        self.in_conditional = False

    def analyze(
        self,
        block: SLRBlock,
        variable_context: dict[str, Any],
    ) -> DataFlowAnalysis:
        """Analyze data flow in a block.

        Args:
            block: The SLR block to analyze
            variable_context: Context of variables (QReg, CReg, etc.)

        Returns:
            DataFlowAnalysis containing all data flow information
        """
        analysis = DataFlowAnalysis()
        self.position_counter = 0
        self.in_conditional = False

        # Analyze all operations
        if hasattr(block, "ops"):
            for op in block.ops:
                self._analyze_operation(op, analysis, variable_context)
                self.position_counter += 1

        return analysis

    def _analyze_operation(
        self,
        op: Any,
        analysis: DataFlowAnalysis,
        variable_context: dict[str, Any],
    ) -> None:
        """Analyze a single operation."""
        op_type = type(op).__name__

        if op_type == "Measure":
            self._analyze_measurement(op, analysis)
        elif op_type == "If":
            self._analyze_if_block(op, analysis, variable_context)
        elif hasattr(op, "qargs"):
            # Check if this is a preparation operation
            if self._is_preparation(op):
                self._analyze_preparation(op, analysis)
            else:
                self._analyze_quantum_operation(op, analysis)
        elif hasattr(op, "ops"):
            # Nested block - recurse
            for nested_op in op.ops:
                self._analyze_operation(nested_op, analysis, variable_context)

    def _is_preparation(self, op: Any) -> bool:
        """Check if an operation is a preparation/reset."""
        op_name = type(op).__name__
        return op_name in ["Prep", "Init", "Reset"]

    def _analyze_measurement(self, meas: Any, analysis: DataFlowAnalysis) -> None:
        """Analyze a measurement operation."""
        # Get classical targets
        classical_targets = []
        if hasattr(meas, "cout") and meas.cout:
            classical_targets.extend(
                (cout.reg.sym, cout.index)
                for cout in meas.cout
                if hasattr(cout, "reg") and hasattr(cout.reg, "sym") and hasattr(cout, "index")
            )

        # Analyze quantum sources
        if hasattr(meas, "qargs") and meas.qargs:
            for i, qarg in enumerate(meas.qargs):
                # Individual element measurement
                if hasattr(qarg, "reg") and hasattr(qarg.reg, "sym") and hasattr(qarg, "index"):
                    array_name = qarg.reg.sym
                    index = qarg.index

                    # Get corresponding classical target if exists
                    classical_array = None
                    classical_index = None
                    if i < len(classical_targets):
                        classical_array, classical_index = classical_targets[i]

                    analysis.add_measurement(
                        quantum_array=array_name,
                        quantum_index=index,
                        position=self.position_counter,
                        classical_array=classical_array,
                        classical_index=classical_index,
                    )

    def _analyze_preparation(self, op: Any, analysis: DataFlowAnalysis) -> None:
        """Analyze a preparation/reset operation."""
        if hasattr(op, "qargs") and op.qargs:
            for qarg in op.qargs:
                if hasattr(qarg, "reg") and hasattr(qarg.reg, "sym") and hasattr(qarg, "index"):
                    array_name = qarg.reg.sym
                    index = qarg.index
                    analysis.add_preparation(array_name, index, self.position_counter)

    def _analyze_quantum_operation(self, op: Any, analysis: DataFlowAnalysis) -> None:
        """Analyze a quantum gate operation."""
        if hasattr(op, "qargs") and op.qargs:
            for qarg in op.qargs:
                if hasattr(qarg, "reg") and hasattr(qarg.reg, "sym") and hasattr(qarg, "index"):
                    array_name = qarg.reg.sym
                    index = qarg.index

                    if self.in_conditional:
                        analysis.add_conditional_use(
                            array_name,
                            index,
                            self.position_counter,
                            is_classical=False,
                        )
                    else:
                        analysis.add_gate_use(array_name, index, self.position_counter)

    def _analyze_if_block(
        self,
        if_block: Any,
        analysis: DataFlowAnalysis,
        variable_context: dict[str, Any],
    ) -> None:
        """Analyze an if block."""
        prev_conditional = self.in_conditional
        self.in_conditional = True

        # Analyze condition
        if hasattr(if_block, "cond"):
            self._analyze_condition(if_block.cond, analysis)

        # Analyze then block
        if hasattr(if_block, "ops"):
            for op in if_block.ops:
                self._analyze_operation(op, analysis, variable_context)

        # Analyze else block
        if hasattr(if_block, "else_block") and if_block.else_block and hasattr(if_block.else_block, "ops"):
            for op in if_block.else_block.ops:
                self._analyze_operation(op, analysis, variable_context)

        self.in_conditional = prev_conditional

    def _analyze_condition(self, cond: Any, analysis: DataFlowAnalysis) -> None:
        """Analyze a condition expression."""
        cond_type = type(cond).__name__

        if cond_type == "Bit" and hasattr(cond, "reg") and hasattr(cond.reg, "sym") and hasattr(cond, "index"):
            array_name = cond.reg.sym
            index = cond.index
            analysis.add_conditional_use(
                array_name,
                index,
                self.position_counter,
                is_classical=True,
            )

        # Handle compound conditions
        if hasattr(cond, "left"):
            self._analyze_condition(cond.left, analysis)
        if hasattr(cond, "right"):
            self._analyze_condition(cond.right, analysis)



"""Generic stabilizer check implementations.

This module provides generic implementations for stabilizer checks in
quantum error correction, including syndrome extraction circuits and
check operations that can be used across different QEC codes.
"""

# Copyright 2024 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

from typing import TYPE_CHECKING

from pecos.slr import Barrier, Block, Comment
from pecos.slr.qeclib.qubit import CX, CY, CZ, H, Measure, Prep

if TYPE_CHECKING:
    from pecos.slr import Bit, Qubit
    from pecos.slr.qeclib.qubit.qgate_base import QGate


class Check(Block):
    """Generic stabilizer check operation.

    This class implements a generic stabilizer check operation that applies
    a sequence of Pauli operators to data qubits controlled by an ancilla qubit.
    """

    def __init__(
        self,
        d: list[Qubit],
        paulis: str,
        a: Qubit,
        out: Bit,
        *,
        with_barriers: bool = False,
    ) -> None:
        """Initialize a stabilizer check measurement.

        Args:
            d: List of data qubits to check.
            paulis: String of Pauli operators (X, Y, or Z) to apply to each data qubit.
                Can be a single character (applied to all qubits) or one character per qubit.
            a: Ancilla qubit used for the check measurement.
            out: Classical bit to store the measurement result.
            with_barriers: Whether to insert barrier instructions between operations to prevent
                gate reordering. Defaults to False.

        Raises:
            Exception: If check weight is less than 2.
            Exception: If number of Paulis doesn't match number of data qubits.
            Exception: If invalid Pauli operator is specified.
        """
        super().__init__()

        n: int = len(d)

        if n <= 1:
            msg = "Check must be of weight 2 or more"
            raise Exception(msg)

        if len(paulis) == 1:
            paulis *= n
        elif n != len(paulis):
            msg = "Must use a single Pauli or have the same number of Paulis as data qubits."
            raise Exception(msg)

        for p in paulis:
            if p not in {"X", "Y", "Z"}:
                msg = 'Only "X", "Y" and "Z" are accepted.'
                raise Exception(msg)

        ps = paulis

        self.extend(
            Comment(f"Measure check {ps}"),
            Prep(a),
            H(a),
        )

        for i in range(n):
            if with_barriers:
                self.extend(
                    Barrier(a, d[i]),  # to preserve order
                )

            self.extend(
                self.cp(ps[i], a, d[i]),
            )

            if with_barriers:
                self.extend(
                    Barrier(a, d[i]),  # to preserve order
                )

        self.extend(
            H(a),
            Measure(a) > out,
        )

    @staticmethod
    def cp(p: str, a: Qubit, d: Qubit) -> QGate:
        """Create controlled Pauli gate based on string identifier.

        Args:
            p: Pauli gate identifier ('X', 'Y', or 'Z').
            a: Control qubit.
            d: Target qubit.

        Returns:
            Corresponding controlled Pauli gate.
        """
        if p == "X":
            return CX(a, d)
        if p == "Y":
            return CY(a, d)
        if p == "Z":
            return CZ(a, d)
        msg = f"Symbol '{p}' not supported!"
        raise Exception(msg)
